def main():
    # TODO: Put entry-point code here!
    print("No user code have been added to that project yet")
    print(
"To start some of built-in liteServers:\n"\
"python -m liteserver.device.litePeakSimulator\n"\
"python -m liteserver.device.liteScaler\n"\
"python -m liteserver.device.liteLabjack\n"\
"etc..."\
)
if __name__ == "__main__":
    main()



"""litePet configuration of the EPICS simulated oscilloscope""" 
__version__ = 'v04 2024-04-14'# scope1::Waveform_RBV

_Namespace = 'PVA'

# Local variables, they can be overridden in the command line (option --macro).
D = 'simScope1:'
#``````````````````Definitions````````````````````````````````````````````````
# python expressions and functions, used in the spreadsheet

def slider(minValue,maxValue):
    """Definition of the GUI element: horizontal slider with flexible range"""
    return {'widget':'hslider','opLimits':[minValue,maxValue]}

MediumFont = {'color':3*[220], 'font': ['Arial',12]}
_ = ''

#``````````````````Page attributes, optional``````````````````````````````````
#_Page = {'editable': False,}
#``````````````````Column attributes, optional````````````````````````````````
_Columns = {
1:{'width':140,'justify':'right'},
2:{'width':80},
3:{'width':70},
4:{'width':30},
5:{'width':480},
}
#``````````````````Configuration of rows``````````````````````````````````````
_Rows = [
[{'ATTRIBUTES':MediumFont},
  {f'Parameters of {D}':{'span':[3,1],'justify':'center'}},_,_,
  {D+'VoltOffset':{'span':[1,16],'widget':'vslider','opLimits':[-10,10]}},
  {_:{'embed':f'python -m pvplot P:{D}Waveform_RBV','span':[1,16]}}],
['Run:',{D+'Run':{'attr':'WE'}}, {D+'threads':{'attr':'R'}}],
#['VoltsPerDivSelect',{D+'VoltsPerDivSelect':{'attr':'R'}}],
['VoltsPerDivSelect:',{D+'VoltsPerDivSelect':{'attr':'WE'}}],
['TimePerDivSelect:',{D+'TimePerDivSelect':{'attr':'WE','span':[2,1]}},''],
['VertGainSelect:',{D+'VertGainSelect':{'attr':'WE'}}],
['VoltOffset:',{D+'VoltOffset':{'attr':'WE'}}],
['TriggerDelay:',{D+'TriggerDelay':{'attr':'WE',**slider(0,10)}},
    D+'TriggerDelay'],
['NoiseAmplitude:',{D+'NoiseAmplitude':{'attr':'WE',**slider(0,10)}},
    D+'NoiseAmplitude'],
['UpdateTime:',{D+'UpdateTime':{'attr':'WE',**slider(0.1,10)}},D+'UpdateTime'],
['Waveform:',{D+'Waveform_RBV':{'attr':'R'}}],
['TimeBase:',{D+'TimeBase_RBV':{'attr':'R'}}],
['MaxPoints:',{D+'MaxPoints_RBV':{'attr':'R'}}],
['MinValue:',{D+'MinValue_RBV':{'attr':'R'}}],
['MaxValue:',{D+'MaxValue_RBV':{'attr':'R'}}],
['MeanValue:',{D+'MeanValue_RBV':{'attr':'R'}}],
]



"""litePet configuration of the EPICS simulated oscilloscope""" 
__version__ = 'v04 2024-04-14'# scope1::Waveform_RBV

_Namespace = 'EPICS'

# Local variables, they can be overridden in the command line (option --macro).
D = 'testAPD:scope1:'
#``````````````````Definitions````````````````````````````````````````````````
# python expressions and functions, used in the spreadsheet

def slider(minValue,maxValue):
    """Definition of the GUI element: horizontal slider with flexible range"""
    return {'widget':'hslider','opLimits':[minValue,maxValue]}

MediumFont = {'color':3*[220], 'font': ['Arial',12]}
_ = ''

#``````````````````Page attributes, optional``````````````````````````````````
#_Page = {'editable': False,}
#``````````````````Column attributes, optional````````````````````````````````
_Columns = {
1:{'width':140,'justify':'right'},
2:{'width':80},
3:{'width':30},
4:{'width':30},
5:{'width':480},
}
#``````````````````Configuration of rows``````````````````````````````````````
_Rows = [
[{'ATTRIBUTES':MediumFont},
  {f'Parameters of {D}':{'span':[3,1],'justify':'center'}},_,_,
  {D+'VoltOffset':{'span':[1,16],'widget':'vslider','opLimits':[-10,10]}},
  {_:{'embed':'python -m pvplot E:testAPD:scope1:Waveform_RBV','span':[1,16]}}],
['Run:',{D+'Run':{'attr':'WE'}}],
['VoltsPerDivSelect',{D+'VoltsPerDivSelect_RBV':{'attr':'R'}}],
['TimePerDivSelect',{D+'TimePerDivSelect':{'attr':'WE','span':[2,1]}},''],
['VertGainSelect',{D+'VertGainSelect':{'attr':'WE'}}],
['VoltsPerDivSelect',{D+'VoltsPerDivSelect':{'attr':'WE'}}],
['VoltOffset',D+'VoltOffset_RBV'],
['TriggerDelay',{D+'TriggerDelay':{'attr':'WE',**slider(0,10)}},
    D+'TriggerDelay'],
['NoiseAmplitude',{D+'NoiseAmplitude':{'attr':'WE',**slider(0,2)}},
    D+'NoiseAmplitude'],
['UpdateTime',{D+'UpdateTime':{'attr':'WE',**slider(0.1,10)}},D+'UpdateTime'],
['Waveform',{D+'Waveform_RBV':{'attr':'R'}}],
['TimeBase',{D+'TimeBase_RBV':{'attr':'R'}}],
['MaxPoints',{D+'MaxPoints_RBV':{'attr':'R'}}],
['MinValue',{D+'MinValue_RBV':{'attr':'R'}}],
['MaxValue',{D+'MaxValue_RBV':{'attr':'R'}}],
['MeanValue',{D+'MeanValue_RBV':{'attr':'R'}}],
]



#!/usr/bin/env python3
"""Spreadsheet view of process variables from EPICS or liteServer infrastructure"""
__version__= 'v1.0.1 2025-01-06'# 
import argparse
from . import pypet

pypet.AppName = 'pypeto'
pypet.DefaultNamespace = 'EPICS'

def main():
    parser = argparse.ArgumentParser(description = __doc__
    ,formatter_class=argparse.ArgumentDefaultsHelpFormatter
    ,epilog=f'{pypet.AppName}: {pypet.__version__}')
    parser.add_argument('-a','--access', default=pypet.DefaultNamespace, help=\
     'Infrastructure', choices=['EPICS', 'PVA', 'LITE'])
    parser.add_argument('-c','--configDir',
     default = '/operations/app_store/pypet/pages/', help=\
     f'Config directory')
    parser.add_argument('-e','--dont_embed', action='store_true', help=\
     'Do not embed other applications') 
    parser.add_argument('-f','--file', help='Spreadsheet description page (python file with suffix _pp.py')
    parser.add_argument('-g','--geometry', help=\
    'Relative position (x,y) of the window on the screen, e.g. -g0.2,0.5')
    parser.add_argument('-H','--hidemenubar',  action='store_true',  help=\
    'Hide menuBar and statusBar')
    parser.add_argument('-i','--instance', default='', help=\
    'This argument will be available to target pypage as "builtins.pypage_INSTANCE"')
    parser.add_argument('-r','--restore', action='store_true', help=\
    'Restore the parameter setting from a snapshot')
    parser.add_argument('-R','--readonly', action='store_true', help=\
    'Read only mode: modification of parameters is prohibited')
    parser.add_argument('-s','--server', action='store_true', help=\
    'show server variables')
    parser.add_argument('-v', '--verbose', action='count', default=0, help=\
      'Show more log messages (-vv: show even more).')
    parser.add_argument('device', nargs='?', help=\
     ('device (e.g. simple.test), if specified, then a '\
     'temporary configuration will be build'))
    parser.add_argument('-z', '--zoomin', help=\
      'Zoom the application window by a factor, factor must be >= 1')
    pargs = parser.parse_args()
    if pargs.file is not None and pargs.device is not None:
        printe('--file option is not allowed when a device argument is supplied.')
        sys.exit(0)
    pypet.pargs = pargs

    pypet.run()

    print('Application exit')

if __name__ == "__main__":
    main()




#!/usr/bin/env python3
"""Plotting package for EPICS PVs, ADO and LITE parameters.
"""
__version__ = 'v1.2.3 2024-07-29'# fix: process configuration file
#TODO: if backend times out the gui is not responsive
#TODO: move Add Dataset to Dataset options
#TODO: add dataset arithmetics

import sys, os, time, datetime
timer = time.perf_counter
import numpy as np
from qtpy import QtWidgets as QW, QtGui, QtCore
from qtpy.QtWidgets import QApplication, QMainWindow, QGridLayout
import pyqtgraph as pg
from pyqtgraph.graphicsItems.ViewBox.ViewBoxMenu import ViewBoxMenu
from pyqtgraph import dockarea
from functools import partial
from collections import deque
import traceback
from importlib import import_module

#````````````````````````````Constants````````````````````````````````````````
X,Y = 0,1
Scale,Units = 0,1
#````````````````````````````Helper methods```````````````````````````````````
def printTime(): return time.strftime("%m%d:%H%M%S")
def printi(msg): print((f'INFO_PVP@{printTime()}: '+msg))
def printw(msg): print((f'WARN_PVP@{printTime()}: '+msg))
def printe(msg): print((f'ERR_PVP@{printTime()}: '+msg))
def _printv(msg, level=0):
    if PVPlot.pargs.verbose > level:
        print((f'DBG{level}_PVP@{printTime()}: '+msg))
def printv(msg):   _printv(msg, 0)
def printvv(msg):  _printv(msg, 1)

def croppedText(txt, limit=200):
    if len(txt) > limit:
        txt = txt[:limit]+'...'
    return txt
def prettyDict(rDict, lineLimit=75):
    r=''
    for dev,devVals in rDict.items():
        r += dev+'\n'
        for par, parVals in devVals.items():
            r += '  '+par+':'
            if isinstance(parVals,dict):
                r += '\n'
                for attr, attrVal in parVals.items():
                    r += croppedText(f'    {attr}: {attrVal}',lineLimit)+'\n'
            else:
                r += croppedText(f' {parVals}',lineLimit)+'\n'
    return r

try:    from cad_epics import epics as EPICSAccess
except Exception as e:
    EPICSAccess = None 
    printw(f'EPICS devices are not supported on this host: {e}')
try:
    import liteaccess as liteAccess 
    LITEAccess = liteAccess.Access
    print(f'liteAccess {liteAccess.__version__}')
except Exception as e:
    printw(f'LITE devices are not supported on this host: {e}')
    LITEAccess = None
try:    
    from cad_io import adoaccess
    ADOAccess = adoaccess.IORequest()
except Exception as e:
    printw(f'ADO devices are not supported on this host: {e}')
    ADOAccess = None

def cprint(msg): 
    print('cprint:'+msg)

def get_pv(adopar:str, prop='value'):
    #print(f'>get_pv {adopar}')
    adopar, vslice = split_slice(adopar)
    access = PVPlot.access.get(adopar[:2], (ADOAccess,0))
    access,prefixLength = PVPlot.access.get(adopar[:2], (ADOAccess,0))
    if access is None:
        printe(f'No access method for `{adopar}`')
        sys.exit(1)

    pvTuple = tuple(adopar[prefixLength:].rsplit(':',1))
    rd = access.get(pvTuple)[pvTuple]
    val = rd['value']
    try:
        shape = val.shape
        if len(shape) > 2:
            printe(f'2+dimensional arrays not supported for {dev,par}')
            return None
    except:
        # val does not have attribute shape
        pass
    try:
        ts = rd['timestamp']# EPICS and LITE
    except: # ADO
        ts = rd['timestampSeconds'] + rd['timestampNanoSeconds']*1.e-9

    #printv(f"get_pv {adopar}: {rd['value']} {vslice}")
    if vslice is not None:
        val = val[vslice[0]:vslice[1]]
    return val, ts

def change_plotOption(curveName,color=None,width=None,symbolSize=None,scolor=None):
    dataset = MapOfDatasets.dtsDict[curveName]
    if color != None:
        prop = 'color'
        dataset.pen.setColor(color)
    if width != None:
        prop = 'width'
        dataset.width = width
        dataset.pen.setWidth(width)
    elif symbolSize!=None:
        dataset.symbolSize = symbolSize
    elif scolor!=None:
        dataset.symbolBrush = scolor
    else: return
    try:
        dataset.plotItem.setPen(
          dataset.pen)
    except: cprint('could not set '+prop+' for '+str(curveName))

def split_slice(parNameSlice):
    """Decode 'name[n1:n2]' to 'name',[n1:n2]"""
    devParSlice = parNameSlice.split('[',1)
    if len(devParSlice) < 2:
        return devParSlice[0], None
    sliceStr = devParSlice[1].replace(']','')
    vrange = sliceStr.split(':',1)
    r0 = int(vrange[0])
    if len(vrange) == 1:
        vslice = (r0, r0+1)
    else:
        vslice = (r0, int(vrange[1]))
    return devParSlice[0], vslice
#,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,
def update_data():
    """ called on QtCore.qTimer() event to update plots."""
    tstart = timer()
    tt = round(time.time(),6)
    if tt > PVPlot.lastPlotTime + PVPlot.minPlottingPeriod:
        time2plot = True
        PVPlot.lastPlotTime = tt + PVPlot.minPlottingPeriod
    else:
        time2plot = False

    for dataset in MapOfDatasets.dtsDict.values():
        dataset.update_plot(time2plot)
        if dataset.dataChanged:
            printvv(f'updating {dataset.adoPars}')
            if PVPlot.statisticsWindow is None:
                continue

    if PVPlot.perfmon:
        v = timer()-tstart
        print('update time:'+str(timer()-tstart))

def get_statistics():
    statistics = {}
    for dataset in MapOfDatasets.dtsDict.values():
        s = dataset.get_statistics()
        if len(s) != 0:
            statistics.update(s)
    txt = 'Parm,\tRange,\tMean,\t\tSTD\n'
    for key,v in statistics.items():
        txt += f'{key},\t{v["xrange"]},\t{v["mean"]},\t{v["std"]}\n'
    return txt

def set_legend(dockNum:int, state:bool):
    if state: # legend enabled
        printv(f'add legends to dock{dockNum}')
        widget = PVPlot.mapOfPlotWidgets[dockNum]
        listOfItems = widget.getPlotItem().listDataItems()
        l = pg.LegendItem((100,60), offset=(70,30))  # args are (size, offset)
        l.setParentItem(widget.graphicsItem())
        PVPlot.legend[dockNum] = l
        for item in listOfItems:
            iname = item.name()
            printv(f'set_legend {iname} in dock{dockNum}')
            l.addItem(item, iname)
    else: # legend disabled
        printv(f'remove legend from dock{dockNum}')
        try:    
            PVPlot.legend[dockNum].scene().removeItem(PVPlot.legend[dockNum])
            del PVPlot.legend[dockNum]
        except Exception as e:
            printe('failed to remove legend '+dockNum+':'+str(e))
#,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,
#```````````````````````````DateAxis class: time scale for bottom plot scale``
class DateAxis(pg.AxisItem):
    def tickStrings(self, values, scale, spacing):
        strns = []
        if len(values) == 0: 
            return ''
        rng = max(values)-min(values)
        #if rng < 120:
        #    return pg.AxisItem.tickStrings(self, values, scale, spacing)
        if rng < 3600*24:
            string = '%H:%M:%S'
        elif rng >= 3600*24 and rng < 3600*24*30:
            string = '%d'
        elif rng >= 3600*24*30 and rng < 3600*24*30*24:
            string = '%b'
        elif rng >=3600*24*30*24:
            string = '%Y'
        for x in values:
            try:
                strns.append(time.strftime(string, time.localtime(x)))
            except ValueError:  ## Windows can't handle dates before 1970
                strns.append('')
        return strns
#,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,
#````````````````````````````Datasets`````````````````````````````````````````
class Dataset():
    """ dataset storage, keeps everything what is necessary to plot the curve.
    """
    def __init__(self, dockNum:int, name, paramAndCount):
        self.dockNum = dockNum
        self.name = name
        self.adoPars = paramAndCount # list of max 2 of (adoPar,count)
        self.plotItem = None # plotting object PlotDataItem
        self.pen = None # current pen
        self.width = 1 # pen width
        self.lastTimePlotted = 0.
        self.lastTimeUpdated = 0.
        self.plotWidget = None
        self.viewBox = None
        # members for stripchart plot
        self.viewXSize = 9E9# Wed Mar 14 2255 16:00:00
        self.viewXRange = [0., 9E9]
        self.scrolling = False
        
        # plotting options, described in 
        # http://www.pyqtgraph.org/documentation/graphicsItems/plotdataitem.html#pyqtgraph.PlotDataItem
        self.connect = None
        self.shadowPen = None
        self.fillLevel = None
        self.fillBrush = None
        self.stepMode = None
        self.symbol = None
        self.symbolPen = None
        self.symbolBrush = None
        self.symbolSize = None
        self.pxMode = None
        self.timeAxis = False
        self.statistics = {}
        self.enabled = True

        # ``````````````````` Add plotItem ``````````````````````````````````````
        #dockNum = name.split('.')[0]
        printv('plotItem for: '+str([s for s in self.adoPars])+', name:'+str(dockNum))
        initialWidth = 16
        self.data = [np.empty(initialWidth),np.empty(initialWidth)]# [X,U] data storage
        self.dataPtr = 0
        count = self.adoPars[0][1] #
        printv(f'adoPars,count: {self.adoPars,count}')
        isCorrelationPlot = len(self.adoPars) >= 2
        
        # assign the plotwidget
        if dockNum in PVPlot.mapOfPlotWidgets:
            # dock already exist, use the existing plotwidget
            self.plotWidget = PVPlot.mapOfPlotWidgets[dockNum]
            self.viewBox = self.plotWidget.getViewBox()
        else:
            # viewbox for new dock need to be created
            # create vewBox with plotwidget
            self.viewBox = CustomViewBox(dockNum, self)
            self.viewBox.setMouseMode(self.viewBox.RectMode)
            self.viewBox.sigRangeChangedManually.connect(self.xrangeChanged)
            printv(f'Creating new dock{dockNum}')
            title = None
            if count == 1 and not isCorrelationPlot:
                self.plotWidget = pg.PlotWidget(title=title, viewBox=self.viewBox,
                  axisItems={'bottom':DateAxis(orientation='bottom')})
                self.timeAxis = True
                if dockNum != 0:
                    self.plotWidget.setXLink(PVPlot.mapOfPlotWidgets[0])
            else: 
                self.plotWidget = pg.PlotWidget(title=title, viewBox=self.viewBox)
            #self.plotWidget.showGrid(True,True)
            PVPlot.mapOfPlotWidgets[dockNum] = self.plotWidget
            PVPlot.mapOfDocks[dockNum].addWidget(self.plotWidget)
            printv(f'docks: {PVPlot.mapOfDocks.keys()}, widgets: {PVPlot.mapOfPlotWidgets.keys()}')

            c = PVPlot.config
            if c is None:   xl,yl = 'LabelX','LabelY'
            else:           xl,yl = c.XLABEL,c.YLABEL
            #print(f'd:{self.dockNum}, X,Y: {xl,yl}')
            if self.dockNum == 0 and yl != 'LabelY':
                self.plotWidget.setLabel('left', yl)
            if self.dockNum == 0 and xl != 'LabelX':
                self.plotWidget.setLabel('bottom', xl)
            else:
                if isCorrelationPlot:
                    self.plotWidget.setLabel('bottom',self.adoPars[1][0])
                elif count == 1:
                    self.plotWidget.setLabel('bottom','time', units='date', unitPrefix='')
                else:
                    self.plotWidget.setLabel('bottom',PVPlot.scaleUnits[X][Units])

        # set X and Y ranges
        rangeMap = {X: (PVPlot.pargs.xrange, self.plotWidget.setXRange),
                    Y: (PVPlot.pargs.yrange, self.plotWidget.setYRange)}
        for axis,v in rangeMap.items():
            r,func = v
            if r is None:
                continue
            r = [float(i) for i in v[0].split(':')]
            func(*r)

        # create plotItem with proper pen
        lineNumber = len(PVPlot.mapOfPlotWidgets[self.dockNum].getPlotItem().listDataItems())
        self.pen = pg.mkPen(lineNumber)                
        if self.adoPars[0][0] == '':
            printv('no dataset - no plotItem')
        else:
            self.plotItem = pg.PlotDataItem(name=name, pen=self.pen)

        if self.plotItem:
            self.plotWidget.addItem(self.plotItem)
        self.lastTimePlotted = 0.

    def add_plot(self):
        self.plotWidget.addItem(self.plotItem)

    def remove_plot(self):
        self.plotWidget.removeItem(self.plotItem)

    def __str__(self):
        return f'Dataset {self.name}, x: {self.data[X].shape}'

    def xrangeChanged(self):
        viewRange = self.viewBox.viewRange()
        self.viewXRange = viewRange[X]
        viewlimit = self.viewXRange[1]
        self.viewXSize = viewlimit - self.viewXRange[0]
        self.viewBox.rangestack.append(viewRange)
        self.viewBox.unzooming = False
        #print(f'>rangestack {len(self.viewBox.rangestack)}')
        self.scrolling = (self.viewBox is not None)\
            and viewlimit > self.data[X][self.dataPtr-1]
        #print(f'scrolling: {self.scrolling}')

    def shift_viewXRange(self):
        if self.viewBox.state['autoRange'][X]:
                return
        dx = self.viewXSize*PVPlot.padding
        self.viewXRange[0] += dx
        self.viewXRange[1] += dx
        #print(f'>shift_viewXRange: {self.viewXRange[0], self.viewXRange[1]}')
        self.viewBox.setXRange(self.viewXRange[0], self.viewXRange[1])

    def plot(self, ts):
        # Plot the dataset. It may be several updates between these calls.
        if self.lastTimePlotted == ts:
            return
        printv(f'>plot: {self.name}')
        self.lastTimePlotted = ts
        #print(f'>plot: {self.name, {self.dataPtr}, round(ts,3)}')
        x = self.data[X][:self.dataPtr]
        y = self.data[Y][:self.dataPtr]
        pen = self.pen if self.width else None
        #printvv(f'x:{x}\ny:{y}')
        self.plotItem.setData(x=x, y=y,
            pen = pen,
            #TODO:connect = self.connect,
            #TODO:shadowPen = self.shadowPen,
            #TODO:fillLevel = self.fillLevel,
            #TODO:fillBrush = self.fillBrush,
            #TODO:stepMode = self.stepMode,
            symbol = self.symbol,
            #TODO:symbolPen = self.symbolPen,
            symbolPen = None,
            symbolBrush = self.symbolBrush,
            #TODO:symbolBrush = self.pen.color(),
            symbolSize = self.symbolSize,
            #TODO:pxMode = self.pxMode,
        )

    def update_plot(self, time2plot):
        curvePars = self.adoPars
        yd,ts = None,None
        try:
            #Limitation: only one (first) curve will be plotted 
            yd, ts = get_pv(curvePars[0][0])
        except Exception as e:
            #printv('got '+str((yd,ts))+', from:'+str(curvePars[0][0])+', except:'+str(e))
            printw(f'Exception getting {curvePars[0][0]}: {e}')
            return
        if ts:
            if ts == self.lastTimeUpdated:
                self.dataChanged = False
                printv(f'curve {self.name} did not change {round(ts,3)}')
                #if time2plot:
                #    self.plot(ts)
                return
        printv(f'update_plot: {self.name,curvePars}')
        self.dataChanged = True
        self.lastTimeUpdated = ts
        #printv(f'update_plot: {curvePars}, data:{yd}')
        #print(f'update {self.name, round(ts,3), round(self.lastTimePlotted,3)}')
        try:    
            l = len(yd)
            if l == 1: yd = yd[0]
        except: 
            l = 1

        # Evaluate X and Y arrays
        if l > 1:
            #printv('Array plot')
            self.data[Y] = np.array(yd)
            if len(curvePars) > 1:
                # use last item as horizontal axis
                self.data[X],*_ = get_pv(curvePars[-1][0])
            else:
                # scaled sample number
                self.data[X] = np.arange(len(yd))*PVPlot.scaleUnits[X][Scale]
            self.dataPtr = len(yd)
            self.plot(ts)
        else:
            # the plot is scrolling or correlation plot
            ptr = self.dataPtr
            if ptr >= PVPlot.pargs.limit:
                # do not extent the data buffer, roll it over instead
                self.data[X] = np.roll(self.data[X],-1)
                self.data[Y] = np.roll(self.data[Y],-1)
                ptr -= 1
            self.data[Y][ptr] = yd
            if len(curvePars) > 1:
                #print('Correlation Plot') 
                printv(f'correlation plot: {curvePars[1][0]}')
                try:
                    v,*_ = get_pv(curvePars[1][0])
                    try:    v = v[0]
                    except: pass 
                    self.data[X][ptr] = v
                except Exception as e:
                    printe('no data from '+str(curvePars[1][0]))
            else:
                # scrolling plot with time scale
                #print(f'ts: {round(ts,3), round(self.viewXRange[1],3)}')
                if self.scrolling and ts > self.viewXRange[1]:
                    self.shift_viewXRange()
                self.data[X][ptr] = ts
            ptr += 1
            self.dataPtr = ptr

            # re-alocate arrays,if necessary
            if (ptr >= self.data[Y].shape[0]):
                #print(f'realllocate: ptr: {ptr,self.data[Y].shape[0]}')
                tmp = self.data
                self.data = [np.empty(self.data[Y].shape[0] * 2),
                    np.empty(self.data[Y].shape[0] * 2)]
                self.data[Y][:tmp[Y].shape[0]] = tmp[Y]
                self.data[X][:tmp[X].shape[0]] = tmp[X]
                #printi(f'adjust {self.name} from {tmp[X].shape} to {self.data[X].shape}')

            if time2plot:
                self.plot(ts)

    def _visibleRange(self):
        npt = self.dataPtr
        #print(f'viewRange={self.viewBox.viewRange()}')
        xr = self.viewBox.viewRange()[X]
        for ileft in range(npt):
            if self.data[X][ileft] > xr[0]:
                break
        #print(f'last={self.data[X][npt-2], self.data[X][npt-1]}')
        for i in range(npt):
            if self.data[X][npt-i-1] < xr[1]:
                break
        iright = npt -i
        return ileft,iright

    def show_yProjection(self):
        ileft,iright = self._visibleRange()
        h = np.histogram(self.data[Y][ileft:iright],100)
        PVPlot.yProjectionPlotItem.setData(h[1],h[0],pen='b',stepMode=True)

    def get_statistics(self):
        ileft,iright = self._visibleRange()
        parname = self.adoPars[0][0].rsplit(':',1)[1]
        r = {parname:{}}
        rp = r[parname]
        rp['xrange'] = (ileft,iright)
        rp['mean'] = np.mean(self.data[Y][ileft:iright]).astype('f4')
        rp['std'] = np.std(self.data[Y][ileft:iright]).astype('f4')
        return r

    def get_visibleData(self):
        ileft,iright = self._visibleRange()
        r = (self.data[X][ileft:iright], self.data[Y][ileft:iright])
        return r
        
class MapOfDatasets():
    """Global dictionary of Datasets, provides safe methods to add and remove 
    the datasets"""
    dtsDict = {}    #{curveName:dataset,,,}
    
    def add(dockNum:int, mapOfCurves):
        """add new datasets, the adoPars is the space delimited string of 
        source ado:parameters."""
        printv(f'>MapOfDatasets.add({dockNum, mapOfCurves})')
        for i, kv in enumerate(mapOfCurves.items()):
            curveName,devList = kv
            printv(f'curveName: {curveName,devList}')
            if isinstance(devList,str):
                devList = [devList]
            else:
                devList = list(devList)
            devList.reverse()
            pnameAndCount = []
            for adoPar in devList:
                adoPar = adoPar.lstrip()#remove leading whitespaces
                printv(f'adoPar: {adoPar}')
                if adoPar.startswith('device('):
                    e = adoPar.index(')')
                    rest = adoPar[e+1:]# it could be arithmetics
                    adoPar = adoPar[7:e]
                try:
                    printv(f'check if {adoPar}, is alive')
                    valts = get_pv(adoPar) # check if parameter is alive
                    if valts is None:
                        printw('Could not add {adoPar}')
                        return 2
                except Exception as e:
                    printw(f'Exception in getting parameter {adoPar}: {e}')
                    return 1
                val,ts = valts
                try:    count = len(val)
                except: count = 1
                pnameAndCount.append((adoPar,count))
            printv('adding '+str(pnameAndCount)+' to datasets['+curveName+']')
            MapOfDatasets.dtsDict[curveName] = Dataset(dockNum, curveName, pnameAndCount)
            PVPlot.mapOfCurves[dockNum].append(curveName)
        printv(f'MapOfDatasets: {[(k,v.adoPars) for k ,v in  MapOfDatasets.dtsDict.items()]}')
        return 0
    
    def remove(name):
        printv('MapOfDatasets.remove '+name)
        dataset = MapOfDatasets.dtsDict[name]
        dataset.plotWidget.removeItem(dataset.plotItem)
        del MapOfDatasets.dtsDict[dataset.name]

class PopupWindow(QW.QWidget): 
    """ This "window" is a QWidget. If it has no parent, it
    will appear as a free-floating window as we want.
    Note: The created window will not be closed on app exit.
    To make it happen, the MainWindow should call close this widget in its closeEvent() 
    """
    def __init__(self):
        super().__init__()
        qr = PVPlot.qWin.geometry()
        self.setGeometry(QtCore.QRect(qr.x(), qr.y(), 0, 0))
        self.setWindowTitle('Statistics')
        layout = QW.QVBoxLayout()
        self.label = QW.QLabel()
        self.label.setTextInteractionFlags(QtCore.Qt.TextSelectableByMouse)
        layout.addWidget(self.label)
        self.setLayout(layout)
#,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,
#`````````````````````````````````````````````````````````````````````````````
class CustomViewBox(pg.ViewBox):
    """ defines actions, activated on the right mouse click in the dock
    """
    def __init__(self, dockNum:int, dataset):
        #self.dockNum = kwds['name'] # cannot use name due to an issue in demo
        #del kwds['name'] # the name in ViewBox.init fails in demo
        self.dockNum = dockNum
        self.dataset = dataset# master dataset, it defines horizontal axis
        #print('CustomViewBox: '+str(self.dockNum))

        # call the init method of the parent class
        super(CustomViewBox, self).__init__()
        self.setDefaultPadding(0.)# standard is 0.02

        # IMPORTANT: menu creation is deferred because it is expensive 
        # and often the user will never see the menu anyway.
        self.menu = None
        self.cursors = set()
        self.unzooming = False
        self.rangestack = deque([],10) #stack of 10 of view ranges
           
    #v32#def mouseClickEvent(self, ev) removed, due to blank exports

    def raiseContextMenu(self, ev):
        # Let the scene add on to the end of our context menu
        menuIn = self.getContextMenus()        
        menu = self.scene().addParentContextMenus(self, menuIn, ev)
        menu.popup(ev.screenPos().toPoint())
        return True

    def getContextMenus(self, event=None):
        """ This method will be called when this item's children want to raise
        a context menu that includes their parents' menus.
        """
        if self.menu:
            printv('menu exist')
            return self.menu
        printv('getContextMenus for '+str(self.dockNum))
        self.menu = ViewBoxMenu(self)
        self.menu.setTitle(f'dock{self.dockNum} options..')

        # unzoom last
        unzoom = self.menu.addAction("&UnZoom")
        unzoom.triggered.connect(lambda: self.unzoom())

        # Datasets options dialog
        setDatasets = self.menu.addAction('Datasets &Options')
        setDatasets.triggered.connect(self.changed_datasetOptions)

        _statistics = self.menu.addAction('Show Statistics')
        _statistics.triggered.connect(self.show_statistics)

        _yProjection = self.menu.addAction('Show yProjection')
        _yProjection.triggered.connect(self.show_yProjection)

        cursorMenu = self.menu.addMenu('Add &Cursor')
        for cursor in ['Vertical','Horizontal']:
            action = cursorMenu.addAction(cursor)
            action.triggered.connect(partial(self.cursorAction,cursor))
        
        labelX = QW.QWidgetAction(self.menu)
        self.labelXGui = QW.QLineEdit('LabelX')
        self.labelXGui.returnPressed.connect(
            lambda: self.set_label('bottom',self.labelXGui))
        labelX.setDefaultWidget(self.labelXGui)
        self.menu.addAction(labelX)
        labelY = QW.QWidgetAction(self.menu)
        self.labelYGui = QW.QLineEdit('LabelY')
        self.labelYGui.returnPressed.connect(
            lambda: self.set_label('left',self.labelYGui))
        labelY.setDefaultWidget(self.labelYGui)
        self.menu.addAction(labelY)
                   
        backgroundAction = QW.QWidgetAction(self.menu)
        backgroundGui = QW.QCheckBox('&Black background')
        backgroundGui.stateChanged.connect(
          lambda x: self.setBackgroundColor(\
          'k' if x == QtCore.Qt.Checked else 'w'))
        backgroundAction.setDefaultWidget(backgroundGui)
        self.menu.addAction(backgroundAction)

        legenAction = QW.QWidgetAction(self.menu)
        legendGui = QW.QCheckBox('&Legend')
        legendGui.setChecked(True)
        legendGui.stateChanged.connect(lambda x: self.set_legend(x))
        legenAction.setDefaultWidget(legendGui)
        self.menu.addAction(legenAction)
        
        runAction = QW.QWidgetAction(self.menu)
        runWidget = QW.QCheckBox('&Stop')
        runWidget.setChecked(PVPlot.stopped)
        runWidget.stateChanged.connect(lambda x: self.set_stop(x))
        runAction.setDefaultWidget(runWidget)
        self.menu.addAction(runAction)

        sleepTimeMenu = self.menu.addMenu('Sleep&Time')
        sleepTimeAction = QW.QWidgetAction(sleepTimeMenu)
        sleepTimeWidget = QW.QDoubleSpinBox()
        sleepTimeWidget.setValue(PVPlot.pargs.sleepTime)
        sleepTimeWidget.setRange(0.001,100)
        sleepTimeWidget.setSuffix(' s')
        sleepTimeWidget.setSingleStep(.1)
        sleepTimeWidget.valueChanged.connect(lambda x: self.set_sleepTime(x))
        sleepTimeAction.setDefaultWidget(sleepTimeWidget)
        sleepTimeMenu.addAction(sleepTimeAction)
        return self.menu

    def cursor_text(self, pos, horizontal):
        #print(f'timeAxis = {self.dataset.timeAxis}')
        if (not horizontal) and self.dataset.timeAxis:
            txt = datetime.datetime.fromtimestamp(pos).strftime('%H:%M:%S')
        else:
            txt = f'{pos:.4g}'
        return txt
        
    def cursorAction(self, direction):
        angle = {'Vertical':90, 'Horizontal':0}[direction]
        pwidget = PVPlot.mapOfPlotWidgets[self.dockNum]
        vid = {'Vertical':0, 'Horizontal':1}[direction]
        vr = pwidget.getPlotItem().viewRange()
        #print(f'vid: {vid,vr[vid]}')
        pos = (vr[vid][1] + vr[vid][0])/2.
        pen = pg.mkPen(color='b', width=1, style=QtCore.Qt.DotLine)
        cursor = pg.InfiniteLine(pos=pos, pen=pen, movable=True, angle=angle
        , label=self.cursor_text(pos,vid), labelOpts={'color':(0,0,0)})#,'fill':(0,255,255)})
        cursor.sigPositionChangeFinished.connect(\
        (partial(self.cursorPositionChanged,cursor)))
        self.cursors.add(cursor)
        pwidget.addItem(cursor)
        self.cursorPositionChanged(cursor)

    def cursorPositionChanged(self, cursor):
        pos = cursor.value()
        horizontal = cursor.angle == 0.
        pwidget = PVPlot.mapOfPlotWidgets[self.dockNum]
        viewRange = pwidget.getPlotItem().viewRange()[horizontal]
        if pos > viewRange[1]:
            pwidget.removeItem(cursor)
            self.cursors.remove(cursor)
        else:
            cursor.label.setText(self.cursor_text(pos, horizontal))

    def changed_datasetOptions(self):
        """Dialog Plotting Options"""
        dlg = QW.QDialog()
        dlg.setWindowTitle(f"Dataset of dock{self.dockNum}")
        dlg.setWindowModality(QtCore.Qt.ApplicationModal)
        dlgSize = 500,200
        dlg.setMinimumSize(*dlgSize)
        rowCount,columnCount = 0,7
        tbl = QW.QTableWidget(rowCount, columnCount, dlg)
        tbl.setHorizontalHeaderLabels(
              ['Enb','Name','PV','Color','Width','Symbol','Size'])
        widths = (30,    50, 100,     30,     80,      40,    80)
        for column,width in enumerate(widths):
            tbl.setColumnWidth(column, width)
        tbl.setShowGrid(False)
        tbl.setSizeAdjustPolicy(
            QW.QAbstractScrollArea.AdjustToContents)
        #tbl.resize(*dlgSize)
        tbl.resize()

        for row,curveName in enumerate(PVPlot.mapOfCurves[self.dockNum]):
            tbl.insertRow(row)
            printv(f'curveName:{curveName}')
            dataset = MapOfDatasets.dtsDict[curveName]
            col = 0
            cbox = QW.QCheckBox()
            cbox.setChecked(dataset.enabled)
            #cbox.stateChanged.connect(lambda x: self.enable_dataset(x))
            cbox.stateChanged.connect(self.enable_dataset)
            cbox.setObjectName(curveName)
            tbl.setCellWidget(row, col, cbox)
            col+=1
            adoparName = dataset.adoPars[0][0]
            tbl.setItem(row, col, QW.QTableWidgetItem(curveName))
            col+=1
            printv(f'dataset:{adoparName}')
            tbl.setItem(row, col, QW.QTableWidgetItem(adoparName))
            col+=1
            # color button for line
            colorButton = pg.ColorButton(color=dataset.pen.color())
            colorButton.setObjectName(curveName)
            colorButton.sigColorChanging.connect(lambda x:
              change_plotOption(str(self.sender().objectName()),color=x.color()))
            tbl.setCellWidget(row, col, colorButton)
            col+=1
            # slider for changing the line width
            widthSlider = QW.QSlider()
            widthSlider.setObjectName(curveName)
            widthSlider.setOrientation(QtCore.Qt.Horizontal)
            widthSlider.setMaximum(10)
            widthSlider.setValue(1)
            widthSlider.valueChanged.connect(lambda x:
              change_plotOption(str(self.sender().objectName()),width=x))
            tbl.setCellWidget(row, col, widthSlider)
            col+=1
            # symbol, selected from a comboBox
            self.symbol = QW.QComboBox() # TODO: why self?
            for symbol in ' ostd+x': self.symbol.addItem(symbol)
            self.symbol.setObjectName(curveName)
            self.symbol.currentIndexChanged.connect(self.set_symbol)
            tbl.setCellWidget(row, col, self.symbol)
            col+=1
            # slider for changing the line width
            symbolSizeSlider = QW.QSlider()
            symbolSizeSlider.setObjectName(curveName)
            symbolSizeSlider.setOrientation(QtCore.Qt.Horizontal)
            symbolSizeSlider.setMaximum(10)
            symbolSizeSlider.setValue(1)
            symbolSizeSlider.valueChanged.connect(lambda x:
              change_plotOption(str(self.sender().objectName()),symbolSize=x))
            tbl.setCellWidget(row, col, symbolSizeSlider)
            col+=1
            # color button for symbol
            symbolColorButton = pg.ColorButton(color=dataset.pen.color())
            symbolColorButton.setObjectName(curveName)
            symbolColorButton.sigColorChanging.connect(lambda x:
              change_plotOption(str(self.sender().objectName()),scolor=x.color()))
            tbl.setCellWidget(row, col,symbolColorButton)
        dlg.exec_()

    def set_symbol(self, x):
        """ Change symbol of the scatter plot. The size and color are taken
        from the curve setting"""
        dtsetName = str(self.sender().objectName())
        symbol = str(self.sender().itemText(x))
        printv('set_symbol for '+dtsetName+' to '+symbol)
        dataset = MapOfDatasets.dtsDict[dtsetName]
        if symbol != ' ':
            dataset.symbol = symbol
            if not dataset.symbolSize:
                dataset.symbolSize = 4 # default size
            if not dataset.symbolBrush:
                dataset.symbolBrush = dataset.pen.color() # symbol color = line color
        else:
            # no symbols - remove the scatter plot
            dataset.symbol = None
            pass
            
    def set_label(self,side,labelGui):
        dock,label = self.dockNum,str(labelGui.text())
        printv('changed_label '+side+': '+str((dock,label)))
        PVPlot.mapOfPlotWidgets[dock].setLabel(side,label, units='')
        # it might be useful to return the prompt back:
        #labelGui.setText('LabelX' if side=='bottom' else 'LabelY')

    def set_legend(self, state):
        state = (state==QtCore.Qt.Checked)
        set_legend(self.dockNum, state)

    def set_stop(self, state):
        PVPlot.stop(state == QtCore.Qt.Checked)

    def set_sleepTime(self, itemData):
        #print('setting SleepTime to: '+str(itemData))
        PVPlot.pargs.sleepTime = itemData
        PVPlot.qTimer.stop()
        PVPlot.qTimer.start(int(PVPlot.pargs.sleepTime*1000))

    def unzoom(self):
        #print('>unzoom')
        try:
            if not self.unzooming:
                self.rangestack.pop()
            self.unzooming = True
            viewRange = self.rangestack.pop()
        except IndexError:
            #printw(f'nothing to unzoom')
            self.enableAutoRange()
            return
        #print(f'<rangestack {len(self.rangestack)}')
        self.setRange(xRange=viewRange[X], yRange=viewRange[Y], padding=None,
            update=True, disableAutoRange=True)

    def show_statistics(self):
        if PVPlot.statisticsWindow is None:
            PVPlot.statisticsWindow = PopupWindow()
        txt = get_statistics()
        PVPlot.statisticsWindow.label.setText(txt)
        PVPlot.statisticsWindow.show()
        PVPlot.statistics = {}

    def show_yProjection(self):
        if PVPlot.yProjectionWindow is None:
            PVPlot.yProjectionWindow = pg.PlotWidget()
            PVPlot.yProjectionPlotItem = pg.PlotDataItem()
            PVPlot.yProjectionWindow.addItem(PVPlot.yProjectionPlotItem)

        PVPlot.yProjectionWindow.show()
        PVPlot.yProjectionWindow.setWindowTitle(
            f'Vertical projection of {self.dataset.adoPars[0][0].rsplit(":",1)[1]}')
        self.dataset.show_yProjection()

    def enable_dataset(self, state):
        dtsetName = str(self.sender().objectName())
        print(f'enable_dataset {dtsetName, (state==QtCore.Qt.Checked)}')
        dataset = MapOfDatasets.dtsDict[dtsetName]
        enabled = (state==QtCore.Qt.Checked)
        dataset.enabled = enabled
        dataset.add_plot() if enabled else dataset.remove_plot()
#,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,

def callback(args):
    #print(f'cb: {args}')
    for hostDevPar, pardict in args.items():
        try:
            axis,units = PVPlot.subscribedParMap[(hostDevPar)]
        except Exception as e:
            printw(f'callback exception for {hostDevPar}: {e}')
            #print(f'map: {PVPlot.subscribedParMap}')
            continue
        scale = pardict['value'][0]
        printv(f'axis={axis}, units={units}, scale={scale}')
        PVPlot.scaleUnits[axis][Scale] = scale

def add_curves(dockNum:int, curveMap:str):
    # if dock name is new then create new dock, otherwise extend the 
    # existing one with new curve
    
    #curves = [x for x in MapOfDatasets.dtsDict]
    #docks = [x.split('.')[0] for x in curves]
    curves = MapOfDatasets.dtsDict.keys()
    docks = PVPlot.mapOfDocks.keys()
    printv(f'>addcurves curves {curveMap} to {curves,docks}')
    if dockNum not in docks:
        printv(f'adding new dock{dockNum}')
        PVPlot.mapOfDocks[dockNum] = dockarea.Dock(str(dockNum),
          size=(500,200), hideTitle=True)
        if dockNum == 0:
            PVPlot.dockArea.addDock(PVPlot.mapOfDocks[dockNum], 'right',
            closable=True)
        else:
            PVPlot.dockArea.addDock(PVPlot.mapOfDocks[dockNum], 
              'top', PVPlot.mapOfDocks[0], closable=True) #TODO:closable does not work
        PVPlot.mapOfCurves[dockNum] = []
    if MapOfDatasets.add(dockNum, curveMap):
            printe(f'in add_curves: {dockNum, curveMap}')

def excepthook(exc_type, exc_value, exc_tb):
    """To uncover exceptions, missed in Qt"""
    tb = "".join(traceback.format_exception(exc_type, exc_value, exc_tb))
    print("error caught!:")
    print("error message:\n", tb)
    QApplication.quit()
    # or QApplication.exit(0)

class tmpClass():
    DOCKS = [{}]

def parse_input_parameters(pargs):
    fn = pargs.file
    if fn:
        sys.path.append(pargs.configDir)
        printv(f'importing {fn}')
        try:
            ConfigModule = import_module(fn)
            configFormat = ConfigModule.configFormat
        except ModuleNotFoundError as e:
            printe(f'Trying to import {pargs.configDir}/{fn}: {e}')
            sys.exit(0)
        return ConfigModule
    else:
        for parm in pargs.parms.split():
            parms = (pargs.prefix+parm).split(',')
            if len(parms) == 2:
                parms = (parms[0],pargs.prefix+parms[1])
            tmpClass.DOCKS[0][parm] =  parms
        return tmpClass

class PVPlot():
    config = None
    pargs = None
    mapOfPlotWidgets = {}# active plotItems
    mapOfDocks = {}# dockAreas
    mapOfCurves = {}
    padding = 0.1
    scaleUnits = [[1,'Sample'],[1,'Count']]
    subscribedParMap = {}
    perfmon = False # option for performance monitoring
    legend = {}# unfortunately we have to keep track of legends
    access = {'E:':(EPICSAccess,2), 'L:':(LITEAccess,2)}
    qWin = None
    qTimer = QtCore.QTimer()
    dockArea = None
    minPlottingPeriod = 1/10.# the data will not be plotted faster than that limit.
    lastPlotTime = 0.
    statisticsWindow = None
    yProjectionWindow = None
    yProjectionPlotItem = None
    stopped = False

    def start():
        pargs = PVPlot.pargs
        printv(f'pargs: {PVPlot.pargs}')

        try:    os.environ["QT_SCALE_FACTOR"] = str(pargs.zoomin)
        except: pass
        qApp = QApplication([])
        PVPlot.qWin = QMainWindow()
        PVPlot.dockArea = dockarea.DockArea()
        PVPlot.qWin.setCentralWidget(PVPlot.dockArea)
        PVPlot.qWin.resize(1000,500)
        ## Switch to using white background and black foreground
        pg.setConfigOption('background', 'w')
        pg.setConfigOption('foreground', 'k')

        # Shortcuts
        shortcut = QW.QShortcut(QtGui.QKeySequence("Ctrl+H"), PVPlot.qWin)
        shortcut.activated.connect(PVPlot.actionHelp)
        shortcut = QW.QShortcut(QtGui.QKeySequence("Ctrl+S"), PVPlot.qWin)
        shortcut.activated.connect(partial(PVPlot.actionStop,True))
        shortcut = QW.QShortcut(QtGui.QKeySequence("Ctrl+Q"), PVPlot.qWin)
        shortcut.activated.connect(partial(PVPlot.actionStop,False))
        shortcut = QW.QShortcut(QtGui.QKeySequence("Ctrl+U"), PVPlot.qWin)
        shortcut.activated.connect(PVPlot.actionUnzoom)
        from . import helptxt
        PVPlot.helptxt = helptxt.txt

        #``````````process configuration file
        if pargs.file is not None:
            PVPlot.config = parse_input_parameters(pargs)

            try:    pargs.limit = PVPlot.config.POINTS
            except: pass

            try:    title = PVPlot.config.TITLE
            except: title = title = f'pvplot {pargs.parms}'
            PVPlot.qWin.setWindowTitle(title)
            dockList = PVPlot.config.DOCKS
            for dockNum,curveMap in enumerate(dockList):
                printv(f'add_curves to dock{dockNum}, {curveMap}')
                add_curves(dockNum, curveMap)
        #,,,,,,,,,,
        else:
            if pargs.dock:
                for par in pargs.dock:
                    dockNum = par[0][0]
                    adopar = par[0][1:].lstrip()
                    add_curves(dockNum, {adopar:adopar})
            else:
                # plots for the main dock
                if not ',' in pargs.parms:
                    for parName in pargs.parms.split():
                        add_curves(0, {parName:pargs.prefix+parName})
                else:
                    x,y = pargs.parms.split(',')
                    add_curves(0, {f'{y} vs {x}': (pargs.prefix+x, pargs.prefix+y)})

        if len(MapOfDatasets.dtsDict) == 0:
            printe(f'No datasets created')
            sys.exit(1)

        for dockNum in PVPlot.mapOfDocks:
            set_legend(dockNum, True)

        # Subscriptions. Only LITE system is supported.
        if pargs.xscale is not None:
            print(f'infra: {pargs.prefix}')
            infrastructure = pargs.prefix[:2]
            if infrastructure != 'L:':
                printe(f'The --xscale option is supported only for LITE infrastucture')
                sys.exit(1)
            hostDev = pargs.prefix[2:]
            if hostDev[-1] == ':':
                hostDev = hostDev[:-1]
            par = pargs.xscale
            printv(f'subscribing: {hostDev,par}')
            info = LITEAccess.info((hostDev,par))
            printv(f'info of {hostDev,par}: {info}')
            units = info[par].get('units','')
            PVPlot.scaleUnits[X][1] = units
            LITEAccess.subscribe(callback, (hostDev,par))
            PVPlot.subscribedParMap[(hostDev,par)] = [X, units]

        sys.excepthook = excepthook

        update_data()

        ## Start a timer to rapidly update the plot in pw
        PVPlot.qTimer.timeout.connect(update_data)
        PVPlot.qTimer.start(int(pargs.sleepTime*1000))

        PVPlot.qWin.show()
        PVPlot.qWin.resize(640,480)

        # start GUI
        ret = qApp.instance().exec_()
        print('Application exited')
        sys.exit(ret)

    def stop(state):
        PVPlot.stopped = state
        if state == True:
            PVPlot.qTimer.stop()
        else:
            PVPlot.qTimer.start(int(PVPlot.pargs.sleepTime*1000))

    #``````````Shorthcut handlers
    def actionHelp():
        PVPlot.helpWin = PopupWindow()
        PVPlot.helpWin.label.setText(PVPlot.helptxt)
        PVPlot.helpWin.show()

    def actionStop(state:bool):
        PVPlot.stop(state)

    def actionUnzoom():
        md = PVPlot.mapOfDocks
        docks = md.keys()
        for dockNum in docks:
            vb = PVPlot.mapOfPlotWidgets[dockNum].getPlotItem().getViewBox()
            vb.unzoom()



#!/usr/bin/env python3
"""Plotting package for EPICS PVs (CA and PVA), LITE and ADO parameters.
"""
__version__ = 'v1.6.2 2025-03-01'# Corrected for use with multiple namespaces
#TODO: if backend times out the gui is not responsive
#TODO: move Add Dataset to Dataset options
#TODO: add dataset arithmetics
#TODO: handle XCELLNAME in mon file

import sys, os, time, datetime
timer = time.perf_counter
import numpy as np
from qtpy import QtWidgets as QW, QtGui, QtCore
from qtpy.QtWidgets import QApplication, QMainWindow, QGridLayout
import pyqtgraph as pg
from pyqtgraph.graphicsItems.ViewBox.ViewBoxMenu import ViewBoxMenu
from pyqtgraph import dockarea
from functools import partial
from collections import deque
import traceback
from importlib import import_module

#````````````````````````````Constants````````````````````````````````````````
X,Y = 0,1
Scale,Units = 0,1
TxtColor = (0,128,128)# color of the plot texts
class color:
   PURPLE = '\033[95m'
   CYAN = '\033[96m'
   DARKCYAN = '\033[36m'
   BLUE = '\033[94m'
   GREEN = '\033[92m'
   YELLOW = '\033[93m'
   RED = '\033[91m'
   BOLD = '\033[1m'
   UNDERLINE = '\033[4m'
   END = '\033[0m'
#````````````````````````````Helper methods```````````````````````````````````
def printTime(): return time.strftime("%m%d:%H%M%S")
def printi(msg): print((f'INFO_PVP@{printTime()}: '+msg))
def printw(msg): print((f'WARN_PVP@{printTime()}: '+msg))
def printe(msg): print((f'ERR_PVP@{printTime()}: '+msg))
def _printv(msg, level=0):
    if PVPlot.pargs.verbose > level:
        print((f'DBG{level}_PVP@{printTime()}: '+msg))
def printv(msg):   _printv(msg, 0)
def printvv(msg):  _printv(msg, 1)

def croppedText(txt, limit=200):
    if len(txt) > limit:
        txt = txt[:limit]+'...'
    return txt
def prettyDict(rDict, lineLimit=75):
    r=''
    for dev,devVals in rDict.items():
        r += dev+'\n'
        for par, parVals in devVals.items():
            r += '  '+par+':'
            if isinstance(parVals,dict):
                r += '\n'
                for attr, attrVal in parVals.items():
                    r += croppedText(f'    {attr}: {attrVal}',lineLimit)+'\n'
            else:
                r += croppedText(f' {parVals}',lineLimit)+'\n'
    return r

AccessModule = {}
def check_pv(adopar:str):
    """Check if PV exists, should be called once, prior to any get_pv()"""
    prefix = adopar[:2]
    if prefix[1] != ':':# Set default namespace as ADO
       prefix = 'A:'
    if prefix not in AccessModule:
        if prefix == 'E:':#Check if EPICS caproto interface is available
            try:
                from . import cad_epics as EPICSAccess
            except: 
                try:    from cad_epics import epics as EPICSAccess
                except: pass
            try:
                print(f'EPICS Channels Access interface {EPICSAccess.__version__}')
                AccessModule[prefix] = EPICSAccess
            except:
                printe(f'EPICS PVA devices are not supported on this host')
                sys.exit(1)
        elif prefix == 'V:':# Check if EPICS PVAccess interface is available
            try:
                from . import cad_pvaccess as PVAccess
            except:
                try:    from cad_pvaccess import pvaccess as PVAccess
                except: pass
            try:
                print(f'EPICS PVAccess interface {PVAccess.__version__}')
                AccessModule[prefix] = PVAccess
            except:
                printe(f'EPICS PVA devices are not supported on this host')
                sys.exit(1)
        elif prefix == 'L:':# Check if LiteServer interface is available
            try:
                import liteaccess as liteAccess 
                print(f'liteAccess interface {liteAccess.__version__}')
                AccessModule[prefix] = liteAccess.Access
            except Exception as e:
                printe(f'LITE devices are not supported on this host: {e}')
                sys.exit(1)
        elif prefix == 'A:':# Check if ADO interface is available
            try:
                from cad_io import adoaccess
                ADOAccess = adoaccess.IORequest()
                print(f'ADOAccess interface {ADOAccess.__version__}')
                AccessModule[prefix] = ADOAccess
            except Exception as e:
                printe(f'ADO devices are not supported on this host: {e}')
                sys.exit(1)
        # It is not recommended to plot P2PLant devices.
        #elif prefix == 'P:':# Check if P2Plant interface is available
        #    try:
        #        from p2plantaccess import Access as P2PAccess
        #        print(f'ADOAccess interface {P2PAccess.__version__}')
        #        AccessModule[prefix] = P2PAccess
        #    except Exception as e:
        #        printe(f'P2Plant devices are not supported on this host: {e}')
        #        sys.exit(1)
        else:
            printe(f'Not supported namespace `{prefix}`')
            sys.exit(1)
    return get_pv(adopar)

def get_pv(adopar:str, prop='value'):
    #print(f'>pvp get_pv {adopar}')
    adopar, vslice = split_slice(adopar)
    try:
        access = AccessModule[adopar[:2]]
        adopar = adopar[2:]
    except:
        access = ADOAccess
    if access is None:
        printe(f'No access method for `{adopar}`')
        sys.exit(1)
    pvTuple = tuple(adopar.rsplit(':',1))
    rd = access.get(pvTuple)[pvTuple]
    val = rd['value']
    try:
        shape = val.shape
        if len(shape) > 2:
            printe(f'2+dimensional arrays not supported for {dev,par}')
            return None
    except:
        # val does not have attribute shape
        pass
    try:
        ts = rd['timestamp']# EPICS and LITE
    except: # ADO & PVA
        ts = rd['timestampSeconds'] + rd['timestampNanoSeconds']*1.e-9

    #printv(f"get_pv {adopar}: {rd['value']} {vslice}")
    if vslice is not None:
        val = val[vslice[0]:vslice[1]]
    return val, ts

def change_plotOption(curveName,color=None,width=None,symbolSize=None):
    dataset = MapOfDatasets.dtsDict[curveName]
    if color != None:
        dataset.pen.setColor(color)
    if width != None:
        dataset.width = width
        dataset.pen.setWidth(width)
    if symbolSize!=None:
        dataset.symbolSize = symbolSize
    dataset.plot(time.time())

def split_slice(parNameSlice):
    """Decode 'name[n1:n2]' to 'name',[n1:n2]"""
    devParSlice = parNameSlice.split('[',1)
    if len(devParSlice) < 2:
        return devParSlice[0], None
    sliceStr = devParSlice[1].replace(']','')
    vrange = sliceStr.split(':',1)
    r0 = int(vrange[0])
    if len(vrange) == 1:
        vslice = (r0, r0+1)
    else:
        vslice = (r0, int(vrange[1]))
    return devParSlice[0], vslice

def get_statistics():
    statList = []
    for dataset in MapOfDatasets.dtsDict.values():
        #if dataset.dockNum != PVPlot.statDock:
        #    continue
        s = dataset.get_statistics()
        #print(f's: {s}')
        if len(s) != 0:
            for key,v in s.items():
                statList.append([key, *v["xrange"],
                f'{v["mean"]:.6g}', f'{v["std"]:.6g}',
                f'{v["p2p"]:.6g}'])
    return statList

class PopupHelp(QW.QWidget): 
    """ This "window" is a QWidget. If it has no parent, it
    will appear as a free-floating window as we want.
    Note: The created window will not be closed on app exit.
    To make it happen, the MainWindow should call close this widget in its closeEvent() 
    """
    def __init__(self):
        super().__init__()
        qr = PVPlot.qWin.geometry()
        self.setGeometry(QtCore.QRect(qr.x(), qr.y(), 0, 0))
        self.setWindowTitle('PVPlot short help')
        layout = QW.QVBoxLayout()
        self.label = QW.QLabel()
        self.label.setTextInteractionFlags(QtCore.Qt.TextSelectableByMouse)
        layout.addWidget(self.label)
        self.setLayout(layout)

class TableModel(QtCore.QAbstractTableModel):
    def __init__(self, data, headers):
        super(TableModel, self).__init__()
        self._data = data
        self.headers = headers

    def headerData(self, section: int, orientation: QtCore.Qt.Orientation, role=QtCore.Qt.DisplayRole):
        if role == QtCore.Qt.DisplayRole and orientation == QtCore.Qt.Horizontal:
            try:    r = self.headers[section]
            except: r = f's{section+1}'
            return r

    def data(self, index, role):
        if role == QtCore.Qt.DisplayRole:
            # See below for the nested-list data structure.
            # .row() indexes into the outer list,
            # .column() indexes into the sub-list
            return self._data[index.row()][index.column()]

    def rowCount(self, parent=QtCore.QModelIndex()):
        # The length of the outer list.
        return len(self._data)

    def columnCount(self, parent=QtCore.QModelIndex()):
        # The following takes the first sub-list, and returns
        # the length (only works if all rows are an equal length)
        return len(self._data[0])

    def refresh_data(self, data):
        self._data = data
        #print(f'>refresh_data: {self._data}')
        topLeft = self.createIndex(0, 0)
        bottomRight = self.createIndex(self.rowCount(), self.columnCount())
        self.dataChanged.emit(topLeft, bottomRight)

class TableWindow(QW.QWidget):
    """
    This "window" is a QWidget. If it has no parent, it
    will appear as a free-floating window.
    """
    def __init__(self, data, headers=[]):
        super().__init__()
        self.setWindowTitle('Statistics')
        layout = QW.QVBoxLayout()
        self.table = QW.QTableView()
        self.model = TableModel(data, headers)
        self.table.setModel(self.model)
        hh = self.table.horizontalHeader()       
        #DNW#hh.setSectionResizeMode(QW.QHeaderView.ResizeToContents)
        #hh.setSectionResizeMode(0, QHeaderView.ResizeMode.Stretch)
        hh.setSectionResizeMode(0, QW.QHeaderView.ResizeMode.ResizeToContents)
        hh.setSectionResizeMode(1, QW.QHeaderView.ResizeMode.ResizeToContents)
        hh.setSectionResizeMode(2, QW.QHeaderView.ResizeMode.ResizeToContents)
        layout.addWidget(self.table)
        self.setLayout(layout)

        qr = PVPlot.qWin.geometry()
        '''
        vh = self.table.verticalHeader()
        extra = (30,30)
        sz = (hh.length() + vh.width() + extra[X],
                     vh.length() + hh.height() + extra[Y])
        self.setGeometry(QtCore.QRect(qr.x(), qr.y(), sz[0], sz[1]))
        '''
        self.move(qr.x(), qr.y())
        self.table.setShowGrid(False)
        self.table.setSizeAdjustPolicy(
            QW.QAbstractScrollArea.AdjustToContents)
#,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,
def update_data():
    """ called on QtCore.qTimer() event to update plots."""
    tstart = timer()
    tt = round(time.time(),6)

    for dataset in MapOfDatasets.dtsDict.values():
        dataset.update_plot()
        if dataset.dataChanged:
            if tt > dataset.lastPlotted + PVPlot.minPlottingPeriod:
                dataset.lastPlotted = tt
                dataset.plot(tt)

    if PVPlot.perfmon:
        v = timer()-tstart
        print('update time:'+str(timer()-tstart))

def set_legend(dockNum:int, state:bool):
    if state: # legend enabled
        printv(f'add legends to dock{dockNum}')
        widget = PVPlot.mapOfPlotWidgets[dockNum]
        listOfItems = widget.getPlotItem().listDataItems()
        l = pg.LegendItem((100,60), offset=(70,30), labelTextColor=TxtColor)  # args are (size, offset)
        l.setParentItem(widget.graphicsItem())
        PVPlot.legend[dockNum] = l
        for item in listOfItems:
            iname = item.name()
            printv(f'set_legend {iname} in dock{dockNum}')
            l.addItem(item, iname)
    else: # legend disabled
        printv(f'remove legend from dock{dockNum}')
        try:    
            PVPlot.legend[dockNum].scene().removeItem(PVPlot.legend[dockNum])
            del PVPlot.legend[dockNum]
        except Exception as e:
            printe('failed to remove legend '+dockNum+':'+str(e))
#,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,
#```````````````````````````DateAxis class: time scale for bottom plot scale``
class DateAxis(pg.AxisItem):
    def tickStrings(self, values, scale, spacing):
        strns = []
        if len(values) == 0: 
            return ''
        rng = max(values)-min(values)
        #if rng < 120:
        #    return pg.AxisItem.tickStrings(self, values, scale, spacing)
        if rng < 3600*24:
            string = '%H:%M:%S'
        elif rng >= 3600*24 and rng < 3600*24*30:
            string = '%d'
        elif rng >= 3600*24*30 and rng < 3600*24*30*24:
            string = '%b'
        elif rng >=3600*24*30*24:
            string = '%Y'
        for x in values:
            try:
                strns.append(time.strftime(string, time.localtime(x)))
            except ValueError:  ## Windows can't handle dates before 1970
                strns.append('')
        return strns
#,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,
#````````````````````````````Datasets`````````````````````````````````````````
class Dataset():
    """ dataset storage, keeps everything what is necessary to plot the curve.
    """
    def __init__(self, dockNum:int, name, paramAndCount):
        self.dockNum = dockNum
        self.name = name
        self.adoPars = paramAndCount # list of max 2 of (adoPar,count)
        self.plotItem = None # plotting object PlotDataItem
        self.pen = None # current pen
        self.width = 1 # pen width
        self.lastTimePlotted = 0.
        self.lastTimeUpdated = 0.
        self.plotWidget = None
        self.viewBox = None
        # members for stripchart plot
        self.viewXSize = 9E9# Wed Mar 14 2255 16:00:00
        self.viewXRange = [0., 9E9]
        self.scrolling = False
        
        # plotting options, described in 
        # http://www.pyqtgraph.org/documentation/graphicsItems/plotdataitem.html#pyqtgraph.PlotDataItem
        self.connect = None
        self.shadowPen = None
        self.fillLevel = None
        self.fillBrush = None
        self.stepMode = None
        self.symbol = None
        self.symbolPen = None
        self.symbolBrush = None
        self.symbolSize = None
        self.pxMode = None
        self.timeAxis = False
        self.statistics = {}
        self.enabled = True
        self.lastPlotted = 0

        # ``````````````````` Add plotItem ``````````````````````````````````````
        #dockNum = name.split('.')[0]
        printv('plotItem for: '+str([s for s in self.adoPars])+', name:'+str(dockNum))
        initialWidth = 16
        self.data = [np.empty(initialWidth),np.empty(initialWidth)]# [X,U] data storage
        self.dataPtr = 0
        count = self.adoPars[0][1] #
        printv(f'adoPars,count: {self.adoPars,count}')
        isCorrelationPlot = len(self.adoPars) >= 2
        
        # assign the plotwidget
        if dockNum in PVPlot.mapOfPlotWidgets:
            # dock already exist, use the existing plotwidget
            self.plotWidget = PVPlot.mapOfPlotWidgets[dockNum]
            self.viewBox = self.plotWidget.getViewBox()
        else:
            # viewbox for new dock need to be created
            # create vewBox with plotwidget
            self.viewBox = CustomViewBox(dockNum, self)
            self.viewBox.setMouseMode(self.viewBox.RectMode)
            self.viewBox.sigRangeChangedManually.connect(self.xrangeChanged)
            printv(f'Creating new dock{dockNum}')
            title = None
            if count == 1 and not isCorrelationPlot:
                self.plotWidget = pg.PlotWidget(title=title, viewBox=self.viewBox,
                  axisItems={'bottom':DateAxis(orientation='bottom')})
                self.timeAxis = True
                if dockNum != 0:
                    self.plotWidget.setXLink(PVPlot.mapOfPlotWidgets[0])
            else: 
                self.plotWidget = pg.PlotWidget(title=title, viewBox=self.viewBox)
            #self.plotWidget.showGrid(True,True)
            PVPlot.mapOfPlotWidgets[dockNum] = self.plotWidget
            PVPlot.mapOfDocks[dockNum].addWidget(self.plotWidget)
            printv(f'docks: {PVPlot.mapOfDocks.keys()}, widgets: {PVPlot.mapOfPlotWidgets.keys()}')

            xl,yl = 'LabelX','LabelY'
            try:
                c = PVPlot.config
                xl,yl = c.XLABEL,c.YLABEL
            except: pass
            #print(f'd:{self.dockNum}, X,Y: {xl,yl}')
            if self.dockNum == 0 and yl != 'LabelY':
                self.plotWidget.setLabel('left', yl)
            if self.dockNum == 0 and xl != 'LabelX':
                self.plotWidget.setLabel('bottom', xl)
            else:
                if isCorrelationPlot:
                    self.plotWidget.setLabel('bottom',self.adoPars[1][0])
                elif count == 1:
                    self.plotWidget.setLabel('bottom','time', units='date', unitPrefix='')
                else:
                    self.plotWidget.setLabel('bottom',PVPlot.scaleUnits[X][Units])

        # set X and Y ranges
        rangeMap = {X: (PVPlot.pargs.xrange, self.plotWidget.setXRange),
                    Y: (PVPlot.pargs.yrange, self.plotWidget.setYRange)}
        for axis,v in rangeMap.items():
            r,func = v
            if r is None:
                continue
            r = [float(i) for i in v[0].split(':')]
            func(*r)

        # create plotItem with proper pen
        lineNumber = len(PVPlot.mapOfPlotWidgets[self.dockNum].getPlotItem().listDataItems())
        self.pen = pg.mkPen(lineNumber)                
        if self.adoPars[0][0] == '':
            printv('no dataset - no plotItem')
        else:
            self.plotItem = pg.PlotDataItem(name=name, pen=self.pen)

        if self.plotItem:
            self.plotWidget.addItem(self.plotItem)
        self.lastTimePlotted = 0.

    def add_plot(self):
        self.plotWidget.addItem(self.plotItem)

    def remove_plot(self):
        self.plotWidget.removeItem(self.plotItem)

    def __str__(self):
        return f'Dataset {self.name}, x: {self.data[X].shape}'

    def xrangeChanged(self):
        viewRange = self.viewBox.viewRange()
        self.viewXRange = viewRange[X]
        viewlimit = self.viewXRange[1]
        self.viewXSize = viewlimit - self.viewXRange[0]
        self.viewBox.rangestack.append(viewRange)
        self.viewBox.unzooming = False
        #print(f'>rangestack {len(self.viewBox.rangestack)}')
        self.scrolling = (self.viewBox is not None)\
            and viewlimit > self.data[X][self.dataPtr-1]
        #print(f'scrolling: {self.scrolling}')

    def shift_viewXRange(self):
        if self.viewBox.state['autoRange'][X]:
                return
        dx = self.viewXSize*PVPlot.padding
        self.viewXRange[0] += dx
        self.viewXRange[1] += dx
        #print(f'>shift_viewXRange: {self.viewXRange[0], self.viewXRange[1]}')
        self.viewBox.setXRange(self.viewXRange[0], self.viewXRange[1])

    def plot(self, ts):
        # Plot the dataset. It may be several updates between these calls.
        if self.lastTimePlotted == ts:
            return
        #printv(f'>plot: {self.name}')
        self.lastTimePlotted = ts
        #print(f'>plot: {self.name, {self.dataPtr}, round(ts,3)}')
        x = self.data[X][:self.dataPtr]
        y = self.data[Y][:self.dataPtr]
        pen = self.pen if self.width else None
        #printvv(f'x:{x}\ny:{y}')
        self.plotItem.setData(x=x, y=y,
            pen = pen,
            #TODO:connect = self.connect,
            #TODO:shadowPen = self.shadowPen,
            #TODO:fillLevel = self.fillLevel,
            #TODO:fillBrush = self.fillBrush,
            #TODO:stepMode = self.stepMode,
            symbol = self.symbol,
            #TODO:symbolPen = self.symbolPen,
            symbolPen = None,
            symbolBrush = self.symbolBrush,
            #TODO:symbolBrush = self.pen.color(),
            symbolSize = self.symbolSize,
            #TODO:pxMode = self.pxMode,
        )

    def update_plot(self):
        curvePars = self.adoPars
        yd,ts = None,None
        try:
            #Limitation: only one (first) curve will be plotted 
            yd, ts = get_pv(curvePars[0][0])
        except Exception as e:
            #printv('got '+str((yd,ts))+', from:'+str(curvePars[0][0])+', except:'+str(e))
            printw(f'Exception getting {curvePars[0][0]}: {e}')
            return
        if ts:
            if ts == self.lastTimeUpdated:
                self.dataChanged = False
                #printv(f'curve {self.name} did not change {round(ts,3)}')
                return
        printv(f'update_plot: {self.name,curvePars}')
        self.dataChanged = True
        self.lastTimeUpdated = ts
        #print(f'update {self.name, round(ts,3), round(self.lastTimePlotted,3)}')
        try:    
            l = len(yd)
            if l == 1: yd = yd[0]
        except: 
            l = 1

        # Evaluate X and Y arrays
        if l > 1:
            #printv('Array plot')
            self.data[Y] = np.array(yd)
            if len(curvePars) > 1:
                # use last item as horizontal axis
                self.data[X],*_ = get_pv(curvePars[-1][0])
            else:
                # scaled sample number
                self.data[X] = np.arange(len(yd))*PVPlot.scaleUnits[X][Scale]
            self.dataPtr = len(yd)
            self.plot(ts)
        else:
            # the plot is scrolling or correlation plot
            ptr = self.dataPtr
            if ptr >= PVPlot.pargs.limit:
                # do not extent the data buffer, roll it over instead
                self.data[X] = np.roll(self.data[X],-1)
                self.data[Y] = np.roll(self.data[Y],-1)
                ptr -= 1
            self.data[Y][ptr] = yd
            if len(curvePars) > 1:
                #print('Correlation Plot') 
                printv(f'correlation plot: {curvePars[1][0]}')
                try:
                    v,*_ = get_pv(curvePars[1][0])
                    try:    v = v[0]
                    except: pass 
                    self.data[X][ptr] = v
                except Exception as e:
                    printe('no data from '+str(curvePars[1][0]))
            else:
                # scrolling plot with time scale
                #print(f'ts: {round(ts,3), round(self.viewXRange[1],3)}')
                if self.scrolling and ts > self.viewXRange[1]:
                    self.shift_viewXRange()
                self.data[X][ptr] = ts
            ptr += 1
            self.dataPtr = ptr

            # re-alocate arrays,if necessary
            if (ptr >= self.data[Y].shape[0]):
                #print(f'realllocate: ptr: {ptr,self.data[Y].shape[0]}')
                tmp = self.data
                self.data = [np.empty(self.data[Y].shape[0] * 2),
                    np.empty(self.data[Y].shape[0] * 2)]
                self.data[Y][:tmp[Y].shape[0]] = tmp[Y]
                self.data[X][:tmp[X].shape[0]] = tmp[X]
                #printi(f'adjust {self.name} from {tmp[X].shape} to {self.data[X].shape}')

    def _visibleRange(self):
        npt = self.dataPtr
        #print(f'viewRange={self.viewBox.viewRange()}')
        xr = self.viewBox.viewRange()[X]
        for ileft in range(npt):
            if self.data[X][ileft] > xr[0]:
                break
        #print(f'last={self.data[X][npt-2], self.data[X][npt-1]}')
        for i in range(npt):
            if self.data[X][npt-i-1] < xr[1]:
                break
        iright = npt -i
        return ileft,iright

    def show_yProjection(self):
        ileft,iright = self._visibleRange()
        h = np.histogram(self.data[Y][ileft:iright],100)
        PVPlot.yProjectionPlotItem.setData(h[1],h[0],pen='b',stepMode=True)

    def get_statistics(self):
        ileft,iright = self._visibleRange()
        parname = self.adoPars[0][0]
        r = {parname:{}}
        rp = r[parname]
        rp['xrange'] = (ileft,iright)
        y = np.array(self.data[Y][ileft:iright])
        rp['mean'] = y.mean().astype('f4')
        rp['std'] = y.std().astype('f4')
        rp['p2p'] = (y.max() - y.min()).astype('f4')
        return r

    def get_visibleData(self):
        ileft,iright = self._visibleRange()
        r = (self.data[X][ileft:iright], self.data[Y][ileft:iright])
        return r
        
class MapOfDatasets():
    """Global dictionary of Datasets, provides safe methods to add and remove 
    the datasets"""
    dtsDict = {}    #{curveName:dataset,,,}
    
    def add(dockNum:int, mapOfCurves):
        """add new datasets, the adoPars is the space delimited string of 
        source device:parameters."""
        printv(f'>MapOfDatasets.add({dockNum, mapOfCurves})')
        for i, kv in enumerate(mapOfCurves.items()):
            curveName,devList = kv
            printv(f'curveName: {curveName,devList}')
            if isinstance(devList,str):
                devList = [devList]
            else:
                devList = list(devList)
            devList.reverse()
            pnameAndCount = []
            for adoPar in devList:
                adoPar = adoPar.lstrip()#remove leading whitespaces
                #print(f'adoPar: {adoPar}')
                if adoPar.startswith('device('):
                    e = adoPar.index(')')
                    rest = adoPar[e+1:]# it could be arithmetics
                    adoPar = adoPar[7:e]
                printv(f'check if {adoPar}, is alive')
                if True:#try:
                    valts = check_pv(adoPar) # check if parameter is alive
                    if valts is None:
                        printw('Could not add {adoPar}')
                        return 2
                else:#except Exception as e:
                    printw(f'Exception in getting parameter {adoPar}: {e}')
                    return 1
                val,ts = valts
                try:    count = len(val)
                except: count = 1
                pnameAndCount.append((adoPar,count))
            printv('adding '+str(pnameAndCount)+' to datasets['+curveName+']')
            MapOfDatasets.dtsDict[curveName] = Dataset(dockNum, curveName, pnameAndCount)
            PVPlot.mapOfCurves[dockNum].append(curveName)
        printv(f'MapOfDatasets: {[(k,v.adoPars) for k ,v in  MapOfDatasets.dtsDict.items()]}')
        return 0
    
    def remove(name):
        printv('MapOfDatasets.remove '+name)
        dataset = MapOfDatasets.dtsDict[name]
        dataset.plotWidget.removeItem(dataset.plotItem)
        del MapOfDatasets.dtsDict[dataset.name]
#,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,
#`````````````````````````````````````````````````````````````````````````````
class CustomViewBox(pg.ViewBox):
    """ defines actions, activated on the right mouse click in the dock
    """
    def __init__(self, dockNum:int, dataset):
        #self.dockNum = kwds['name'] # cannot use name due to an issue in demo
        #del kwds['name'] # the name in ViewBox.init fails in demo
        self.dockNum = dockNum
        self.dataset = dataset# master dataset, it defines horizontal axis
        #print('CustomViewBox: '+str(self.dockNum))

        # call the init method of the parent class
        super(CustomViewBox, self).__init__()
        self.setDefaultPadding(0.)# standard is 0.02

        # IMPORTANT: menu creation is deferred because it is expensive 
        # and often the user will never see the menu anyway.
        self.menu = None
        self.cursors = set()
        self.unzooming = False
        self.rangestack = deque([],10) #stack of 10 of view ranges
           
    #v32#def mouseClickEvent(self, ev) removed, due to blank exports

    def raiseContextMenu(self, ev):
        # Let the scene add on to the end of our context menu
        menuIn = self.getContextMenus()        
        menu = self.scene().addParentContextMenus(self, menuIn, ev)
        menu.popup(ev.screenPos().toPoint())
        return True

    def getContextMenus(self, event=None):
        """ This method will be called when this item's children want to raise
        a context menu that includes their parents' menus.
        """
        if self.menu:
            printv('menu exist')
            return self.menu
        printv('getContextMenus for '+str(self.dockNum))
        self.menu = ViewBoxMenu(self)
        self.menu.setTitle(f'dock{self.dockNum} options..')

        # unzoom last
        unzoom = self.menu.addAction("&UnZoom")
        unzoom.triggered.connect(lambda: self.unzoom())

        # Datasets options dialog
        setDatasets = self.menu.addAction('Datasets &Options')
        setDatasets.triggered.connect(self.changed_datasetOptions)

        _statistics = self.menu.addAction('Show &Mean,sigma')
        _statistics.triggered.connect(self.show_statistics)

        _yProjection = self.menu.addAction('Show &YProjection')
        _yProjection.triggered.connect(self.show_yProjection)

        cursorMenu = self.menu.addMenu('Add &Cursor')
        for cursor in ['Vertical','Horizontal']:
            action = cursorMenu.addAction(cursor)
            action.triggered.connect(partial(self.cursorAction,cursor))
        
        # popup help
        phelp = self.menu.addAction("&help")
        phelp.triggered.connect(lambda: PVPlot.actionHelp())

        # Labels
        labelX = QW.QWidgetAction(self.menu)
        self.labelXGui = QW.QLineEdit('LabelX')
        self.labelXGui.returnPressed.connect(
            lambda: self.set_label('bottom',self.labelXGui))
        labelX.setDefaultWidget(self.labelXGui)
        self.menu.addAction(labelX)
        labelY = QW.QWidgetAction(self.menu)
        self.labelYGui = QW.QLineEdit('LabelY')
        self.labelYGui.returnPressed.connect(
            lambda: self.set_label('left',self.labelYGui))
        labelY.setDefaultWidget(self.labelYGui)
        self.menu.addAction(labelY)
                   
        backgroundAction = QW.QWidgetAction(self.menu)
        backgroundGui = QW.QCheckBox('Black background')
        backgroundGui.stateChanged.connect(
          lambda x: self.setBackgroundColor(\
          'k' if x == QtCore.Qt.Checked else 'w'))
        backgroundAction.setDefaultWidget(backgroundGui)
        self.menu.addAction(backgroundAction)

        legenAction = QW.QWidgetAction(self.menu)
        legendGui = QW.QCheckBox('Legend')
        legendGui.setChecked(True)
        legendGui.stateChanged.connect(lambda x: self.set_legend(x))
        legenAction.setDefaultWidget(legendGui)
        self.menu.addAction(legenAction)
        
        runAction = QW.QWidgetAction(self.menu)
        runWidget = QW.QCheckBox('&Stop')
        runWidget.setChecked(PVPlot.stopped)
        runWidget.stateChanged.connect(lambda x: self.set_stop(x))
        runAction.setDefaultWidget(runWidget)
        self.menu.addAction(runAction)

        sleepTimeMenu = self.menu.addMenu('Sleep&Time')
        sleepTimeAction = QW.QWidgetAction(sleepTimeMenu)
        sleepTimeWidget = QW.QDoubleSpinBox()
        sleepTimeWidget.setValue(PVPlot.pargs.sleepTime)
        sleepTimeWidget.setRange(0.001,100)
        sleepTimeWidget.setSuffix(' s')
        sleepTimeWidget.setSingleStep(.1)
        sleepTimeWidget.valueChanged.connect(lambda x: self.set_sleepTime(x))
        sleepTimeAction.setDefaultWidget(sleepTimeWidget)
        sleepTimeMenu.addAction(sleepTimeAction)
        return self.menu

    def cursor_text(self, pos, horizontal):
        #print(f'timeAxis = {self.dataset.timeAxis}')
        if (not horizontal) and self.dataset.timeAxis:
            txt = datetime.datetime.fromtimestamp(pos).strftime('%H:%M:%S')
        else:
            txt = f'{pos:.4g}'
        return txt
        
    def cursorAction(self, direction):
        angle = {'Vertical':90, 'Horizontal':0}[direction]
        pwidget = PVPlot.mapOfPlotWidgets[self.dockNum]
        vid = {'Vertical':0, 'Horizontal':1}[direction]
        vr = pwidget.getPlotItem().viewRange()
        #print(f'vid: {vid,vr[vid]}')
        pos = (vr[vid][1] + vr[vid][0])/2.
        pen = pg.mkPen(color='b', width=1, style=QtCore.Qt.DotLine)
        cursor = pg.InfiniteLine(pos=pos, pen=pen, movable=True, angle=angle
        , label=self.cursor_text(pos,vid), labelOpts={'color':TxtColor})#,'fill':(0,255,255)})
        cursor.sigPositionChangeFinished.connect(\
        (partial(self.cursorPositionChanged,cursor)))
        self.cursors.add(cursor)
        pwidget.addItem(cursor)
        self.cursorPositionChanged(cursor)

    def cursorPositionChanged(self, cursor):
        pos = cursor.value()
        horizontal = cursor.angle == 0.
        pwidget = PVPlot.mapOfPlotWidgets[self.dockNum]
        viewRange = pwidget.getPlotItem().viewRange()[horizontal]
        if pos > viewRange[1]:
            pwidget.removeItem(cursor)
            self.cursors.remove(cursor)
        else:
            cursor.label.setText(self.cursor_text(pos, horizontal))

    def changed_datasetOptions(self):
        """Dialog Plotting Options"""
        dlg = QW.QDialog()
        dlg.setWindowTitle(f"Dataset of dock{self.dockNum}")
        dlg.setWindowModality(QtCore.Qt.ApplicationModal)
        rowCount,columnCount = 0,8
        tbl = QW.QTableWidget(rowCount, columnCount, dlg)
        tbl.setHorizontalHeaderLabels(
                 ['','Name','PV','Color','Width','Symbol','Size',''])
        widths = (10,    50, 100,     20,     80,   35,    80,80)
        for column,width in enumerate(widths):
            tbl.setColumnWidth(column, width)
        tbl.setShowGrid(False)
        tbl.setSizeAdjustPolicy(
            QW.QAbstractScrollArea.AdjustToContents)

        for row,curveName in enumerate(PVPlot.mapOfCurves[self.dockNum]):
            tbl.insertRow(row)
            printv(f'curveName:{curveName}')
            dataset = MapOfDatasets.dtsDict[curveName]
            col = 0
            cbox = QW.QCheckBox()
            cbox.setChecked(dataset.enabled)
            cbox.stateChanged.connect(self.enable_dataset)
            cbox.setObjectName(curveName)
            tbl.setCellWidget(row, col, cbox)

            col+=1
            adoparName = dataset.adoPars[0][0]
            item = QW.QTableWidgetItem(curveName)
            #item.setTextAlignment(QtCore.Qt.AlignRight)
            tbl.setItem(row, col, item)

            col+=1
            printv(f'dataset:{adoparName}')
            txt = adoparName[-24:]
            txt = txt[:12]+'\n'+txt[12:]
            item = QW.QTableWidgetItem(txt)
            tbl.setItem(row, col, QW.QTableWidgetItem(item))

            col+=1
            # color button for line
            colorButton = pg.ColorButton(color=dataset.pen.color())
            colorButton.setObjectName(curveName)
            colorButton.sigColorChanging.connect(lambda x:
              change_plotOption(str(self.sender().objectName()),color=x.color()))
            tbl.setCellWidget(row, col, colorButton)

            col+=1
            # slider for changing the line width
            widthSlider = QW.QSlider()
            widthSlider.setObjectName(curveName)
            widthSlider.setOrientation(QtCore.Qt.Horizontal)
            widthSlider.setMaximum(10)
            widthSlider.setValue(1)
            widthSlider.valueChanged.connect(lambda x:
              change_plotOption(str(self.sender().objectName()),width=x))
            tbl.setCellWidget(row, col, widthSlider)

            col+=1
            # symbol, selected from a comboBox
            self.symbol = QW.QComboBox() # TODO: why self?
            for symbol in ' ostd+x': self.symbol.addItem(symbol)
            self.symbol.setObjectName(curveName)
            self.symbol.currentIndexChanged.connect(self.set_symbol)
            tbl.setCellWidget(row, col, self.symbol)

            col+=1
            # slider for changing symbol size
            symbolSizeSlider = QW.QSlider()
            symbolSizeSlider.setObjectName(curveName)
            symbolSizeSlider.setOrientation(QtCore.Qt.Horizontal)
            symbolSizeSlider.setMaximum(10)
            symbolSizeSlider.setValue(1)
            symbolSizeSlider.valueChanged.connect(lambda x:
              change_plotOption(str(self.sender().objectName()),symbolSize=x))
            tbl.setCellWidget(row, col, symbolSizeSlider)

        #dlg.resize(tbl.width(),tbl.height())
        dialogWidth = tbl.horizontalHeader().length() + 24
        dialogHeight= tbl.verticalHeader().length()   + 24
        dlg.setMinimumSize(dialogWidth, dialogHeight)
        dlg.exec_()

    def set_symbol(self, x):
        """ Change symbol of the scatter plot. The size and color are taken
        from the curve setting"""
        dtsetName = str(self.sender().objectName())
        symbol = str(self.sender().itemText(x))
        dataset = MapOfDatasets.dtsDict[dtsetName]
        #print(f'set_symbol for {dtsetName} to {symbol} {dataset.pen.color()}')
        if symbol != ' ':
            dataset.symbol = symbol
            if not dataset.symbolSize:
                dataset.symbolSize = 4 # default size
            dataset.symbolBrush = dataset.pen.color() # symbol color = line color
            dataset.plot(time.time())#ISSUE: the dataset.plotItem.setSymbol does not work
        else:
            # no symbols - remove the scatter plot
            dataset.symbol = None
            pass
            
    def set_label(self,side,labelGui):
        dock,label = self.dockNum,str(labelGui.text())
        printv('changed_label '+side+': '+str((dock,label)))
        PVPlot.mapOfPlotWidgets[dock].setLabel(side,label, units='')
        # it might be useful to return the prompt back:
        #labelGui.setText('LabelX' if side=='bottom' else 'LabelY')

    def set_legend(self, state):
        state = (state==QtCore.Qt.Checked)
        set_legend(self.dockNum, state)

    def set_stop(self, state):
        PVPlot.stop(state == QtCore.Qt.Checked)

    def set_sleepTime(self, itemData):
        #print('setting SleepTime to: '+str(itemData))
        PVPlot.pargs.sleepTime = itemData
        PVPlot.qTimer.stop()
        PVPlot.qTimer.start(int(PVPlot.pargs.sleepTime*1000))

    def unzoom(self):
        #print('>unzoom')
        try:
            if not self.unzooming:
                self.rangestack.pop()
            self.unzooming = True
            viewRange = self.rangestack.pop()
        except IndexError:
            #printw(f'nothing to unzoom')
            self.enableAutoRange()
            return
        #print(f'<rangestack {len(self.rangestack)}')
        self.setRange(xRange=viewRange[X], yRange=viewRange[Y], padding=None,
            update=True, disableAutoRange=True)

    def show_statistics(self):
        #print(f'print statistics for {self.dockNum}')
        #PVPlot.statDock = self.dockNum
        PVPlot.actionStatistics()

    def show_yProjection(self):
        if PVPlot.yProjectionWindow is None:
            PVPlot.yProjectionWindow = pg.PlotWidget()
            PVPlot.yProjectionPlotItem = pg.PlotDataItem()
            PVPlot.yProjectionWindow.addItem(PVPlot.yProjectionPlotItem)

        PVPlot.yProjectionWindow.show()
        PVPlot.yProjectionWindow.setWindowTitle(
            f'Vertical projection of {self.dataset.name}')
        self.dataset.show_yProjection()

    def enable_dataset(self, state):
        dtsetName = str(self.sender().objectName())
        #print(f'enable_dataset {dtsetName, (state==QtCore.Qt.Checked)}')
        dataset = MapOfDatasets.dtsDict[dtsetName]
        enabled = (state==QtCore.Qt.Checked)
        dataset.enabled = enabled
        dataset.add_plot() if enabled else dataset.remove_plot()
        #set_legend(self.dockNum, enabled)# deletes all legend items
#,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,

def callback(args):
    #print(f'cb: {args}')
    for hostDevPar, pardict in args.items():
        try:
            axis,units = PVPlot.subscribedParMap[(hostDevPar)]
        except Exception as e:
            printw(f'callback exception for {hostDevPar}: {e}')
            #print(f'map: {PVPlot.subscribedParMap}')
            continue
        scale = pardict['value'][0]
        printv(f'axis={axis}, units={units}, scale={scale}')
        PVPlot.scaleUnits[axis][Scale] = scale

def add_curves(dockNum:int, curveMap:str):
    # if dock name is new then create new dock, otherwise extend the 
    # existing one with new curve
    
    #curves = [x for x in MapOfDatasets.dtsDict]
    #docks = [x.split('.')[0] for x in curves]
    curves = MapOfDatasets.dtsDict.keys()
    docks = PVPlot.mapOfDocks.keys()
    printv(f'>addcurves curves {curveMap} to {curves,docks}')
    if dockNum not in docks:
        printv(f'adding new dock{type(dockNum),dockNum}')
        PVPlot.mapOfDocks[dockNum] = dockarea.Dock(str(dockNum),
          size=(500,200), hideTitle=True)
        if dockNum == 0:
            PVPlot.dockArea.addDock(PVPlot.mapOfDocks[dockNum], 'right',
            closable=True)
        else:
            PVPlot.dockArea.addDock(PVPlot.mapOfDocks[dockNum], 
              'top', PVPlot.mapOfDocks[0], closable=True) #TODO:closable does not work
        PVPlot.mapOfCurves[dockNum] = []
    if MapOfDatasets.add(dockNum, curveMap):
            printe(f'in add_curves: {dockNum, curveMap}')

def excepthook(exc_type, exc_value, exc_tb):
    """To uncover exceptions, missed in Qt"""
    tb = "".join(traceback.format_exception(exc_type, exc_value, exc_tb))
    print("error caught!:")
    print("error message:\n", tb)
    QApplication.quit()
    # or QApplication.exit(0)

class tmpClass():
    DOCKS = [{}]

def parse_input_parameters(pargs):
    fn = pargs.file
    if fn:
        sys.path.append(pargs.configDir)
        print(f'Config file: {pargs.configDir}/{fn}')
        try:
            ConfigModule = import_module(fn)
            configFormat = ConfigModule.configFormat
        except ModuleNotFoundError as e:
            printe(f'Trying to import {pargs.configDir}/{fn}: {e}')
            sys.exit(0)
        return ConfigModule
    else:
        for parm in pargs.parms.split():
            parms = (pargs.prefix+parm).split(',')
            if len(parms) == 2:
                parms = (parms[0],pargs.prefix+parms[1])
            tmpClass.DOCKS[0][parm] =  parms
        return tmpClass

class MainWindow(QW.QMainWindow):
    def closeEvent(self,event):
        print('Closing all windows')
        try:    PVPlot.statisticsWindow.close()
        except: pass
        try:    PVPlot.helpWin.close()
        except: pass
        event.accept()

class PVPlot():
    config = None
    pargs = None
    mapOfPlotWidgets = {}# active plotItems
    mapOfDocks = {}# dockAreas
    mapOfCurves = {}
    padding = 0.1
    scaleUnits = [[1,'Sample'],[1,'Count']]
    subscribedParMap = {}
    perfmon = False # option for performance monitoring
    legend = {}# unfortunately we have to keep track of legends
    #access = {'E:':EPICSAccess, 'L:':LITEAccess, 'P:':PVAccess, '2:':P2PAccess}
    qWin = None
    qTimer = QtCore.QTimer()
    dockArea = None
    minPlottingPeriod = 1/10.# the data will not be plotted faster than that limit.
    lastPlotTime = 0.
    statisticsWindow = None
    yProjectionWindow = None
    yProjectionPlotItem = None
    stopped = False
    #statDock = 0

    def start():
        pargs = PVPlot.pargs
        printv(f'pargs: {PVPlot.pargs}')

        try:    os.environ["QT_SCALE_FACTOR"] = str(pargs.zoomin)
        except: pass
        qApp = QApplication([])
        PVPlot.qWin = MainWindow()
        PVPlot.dockArea = dockarea.DockArea()
        PVPlot.qWin.setCentralWidget(PVPlot.dockArea)
        PVPlot.qWin.resize(1000,500)
        ## Switch to using white background and black foreground
        pg.setConfigOption('background', 'w')
        pg.setConfigOption('foreground', 'k')

        # Shortcuts
        shortcut = QW.QShortcut(QtGui.QKeySequence("Ctrl+H"), PVPlot.qWin)
        shortcut.activated.connect(PVPlot.actionHelp)
        shortcut = QW.QShortcut(QtGui.QKeySequence("Ctrl+M"), PVPlot.qWin)
        shortcut.activated.connect(partial(PVPlot.actionStatistics))
        #shortcut = QW.QShortcut(QtGui.QKeySequence("Ctrl+O"), PVPlot.qWin)
        #shortcut.activated.connect(partial(PVPlot.actionDatasetOptions))
        shortcut = QW.QShortcut(QtGui.QKeySequence("Ctrl+S"), PVPlot.qWin)
        shortcut.activated.connect(partial(PVPlot.actionStop))
        shortcut = QW.QShortcut(QtGui.QKeySequence("Ctrl+U"), PVPlot.qWin)
        shortcut.activated.connect(PVPlot.actionUnzoom)
        from . import helptxt
        PVPlot.helptxt = helptxt.txt

        #``````````process configuration file
        if pargs.file is not None:
            PVPlot.config = parse_input_parameters(pargs)

            try:    pargs.limit = PVPlot.config.POINTS
            except: pass

            try:    title = PVPlot.config.TITLE
            except: title = title = f'pvplot {pargs.parms}'
            PVPlot.qWin.setWindowTitle(title)
            dockList = PVPlot.config.DOCKS
            for dockNum,curveMap in enumerate(dockList):
                printv(f'add_curves to dock{dockNum}, {curveMap}')
                add_curves(dockNum, curveMap)
        #,,,,,,,,,,
        else:
            if pargs.dock:
                for par in pargs.dock:
                    dockNum = par[0][0]
                    adopar = par[0][1:].lstrip()
                    add_curves(int(dockNum), {adopar:adopar})
            else:
                # plots for the main dock
                if not ',' in pargs.parms:
                    for parName in pargs.parms.split():
                        add_curves(0, {parName:pargs.prefix+parName})
                else:
                    x,y = pargs.parms.split(',')
                    add_curves(0, {f'{y} vs {x}': (pargs.prefix+x, pargs.prefix+y)})

        if len(MapOfDatasets.dtsDict) == 0:
            printe(f'No datasets created')
            sys.exit(1)

        print(f'Active namespaces: {list(AccessModule.keys())}')
        for dockNum in PVPlot.mapOfDocks:
            set_legend(dockNum, True)

        # Subscriptions. Only LITE system is supported.
        if pargs.xscale is not None:
            print(f'infra: {pargs.prefix}')
            infrastructure = pargs.prefix[:2]
            if infrastructure != 'L:':
                printe(f'The --xscale option is supported only for LITE infrastucture')
                sys.exit(1)
            hostDev = pargs.prefix[2:]
            if hostDev[-1] == ':':
                hostDev = hostDev[:-1]
            par = pargs.xscale
            printv(f'subscribing: {hostDev,par}')
            info = LITEAccess.info((hostDev,par))
            printv(f'info of {hostDev,par}: {info}')
            units = info[par].get('units','')
            PVPlot.scaleUnits[X][1] = units
            LITEAccess.subscribe(callback, (hostDev,par))
            PVPlot.subscribedParMap[(hostDev,par)] = [X, units]

        sys.excepthook = excepthook

        update_data()

        ## Start a timer to rapidly update the plot in pw
        PVPlot.qTimer.timeout.connect(update_data)
        PVPlot.qTimer.start(int(pargs.sleepTime*1000))

        PVPlot.qWin.show()
        PVPlot.qWin.resize(640,480)

        # start GUI
        ret = qApp.instance().exec_()
        print('Application exited')
        sys.exit(ret)

    def stop(state):
        PVPlot.stopped = state
        suffix = ' is stopped'
        if state == True:
            PVPlot.qTimer.stop()
            title = PVPlot.qWin.windowTitle()+suffix
        else:
            PVPlot.qTimer.start(int(PVPlot.pargs.sleepTime*1000))
            title = PVPlot.qWin.windowTitle().replace(suffix,'')
        PVPlot.qWin.setWindowTitle(title)       

    #``````````Shorthcut handlers
    def actionHelp():
        PVPlot.helpWin = PopupHelp()
        PVPlot.helpWin.label.setText(PVPlot.helptxt)
        PVPlot.helpWin.show()

    def actionStop():
        PVPlot.stopped = not PVPlot.stopped
        PVPlot.stop(PVPlot.stopped)

    def actionUnzoom():
        md = PVPlot.mapOfDocks
        docks = md.keys()
        for dockNum in docks:
            vb = PVPlot.mapOfPlotWidgets[dockNum].getPlotItem().getViewBox()
            vb.unzoom()

    def actionStatistics():
        #print(f'>actionStatistics {PVPlot.statDock}')
        sl = get_statistics()
        if PVPlot.statisticsWindow is None:
            header = ['Parameter','XMin','XMax','Mean','StDev','Peak2Peak']
            PVPlot.statisticsWindow = TableWindow(sl, header)
        PVPlot.statisticsWindow.model.refresh_data(sl)
        PVPlot.statisticsWindow.show()        
        



"""Basic EPICS access API using caproto threading.
The API is similar to cad_io.
The devPar argument is tuple (deviceName,parameterName).
The EPICS PV name is concatenation of the deviceName and parameterName

For example: epics.get(('testAPD:scope1:','MaxValue_RBV')) returns 
{('testAPD:scope1:', 'MaxValue_RBV'): {'value': 1.5620877339306638, 'timestamp': 1681488886.923712, 'alarm': 0}}
"""
__version__ = 'v0.0.3 2023-04-14'# No colon separation when splitting PV name to (device,parameter) tuple.

#from time import perf_counter as timer

from caproto.threading.client import Context
_Ctx = Context()

_PVCache = {}# cache of PVs
# the key is EPICS PV name, the fields are as follows:
_PVC_PV = 0# EPICS PV object
_PVC_CB = 1# callback
_PVC_Props = 2# PV properties
_PVC_DevPar = 3# (deviceName, ParameterName)

_Subscriptions = []# list of EPICS PV subscriptions

_CA_data_type_STRING = 14

_dbg = False
def _printd(msg):
    if _dbg:
        print(f'CPAccess:{msg}')

def init():
    return

def _pvname(devPar:tuple):
    return ':'.join(devPar)

def _get_pv(pvName):
    r = _PVCache.get(pvName)
    if r is None:
        #print(f'register pv {pvName}')
        pv,*_ = _Ctx.get_pvs(pvName, timeout=2)
        devPar = tuple(pvName.rsplit(':',1))
        _PVCache[pvName] = [pv, None, None, devPar]
        _fill_PVCacheProps(pv)
    else:
        pv,*_ = r
    return pv

def _fill_PVCacheProps(pv):
    pvName = pv.name
    pvData = pv.read(data_type='time')
    val = pvData.data
    if len(val) == 1:
        try:
            # treat it as numpy
            val = pvData.data[0].item()
        except:
            # data is not numpy
            val = pvData.data[0]
    
    # get properties
    #ISSUE: the caproto reports timestamp as float, the precision for float64 
    #presentation is ~300ns
    featureBit2Letter = {1:'R', 2:'WE'}
    featureCode = pv.access_rights
    features = ''
    for bit, letter in featureBit2Letter.items():
        if bit & featureCode:
            features += letter
    pvControl = pv.read(data_type='control')
    #_printd(f'pvcontrol {pvName}: {pvControl}')
    datatype = pvControl.data_type
    #_printd(f'data_type:{datatype}')
    if datatype == _CA_data_type_STRING:# convert text bytearray to str
        val = val.decode()
    props = {'value':val}
    props['timestamp'] = pvData.metadata.timestamp
    props['count'] = len(pvData.data)
    props['features'] = features
    try:    
        props['units'] = pvControl.metadata.units.decode()
        if props['units'] == '':   props['units'] = None
    except: pass

    try:    props['engLow'] = pvControl.metadata.lower_ctrl_limit
    except: pass

    try:    
        props['engHigh'] = pvControl.metadata.upper_ctrl_limit
        if props['engHigh'] == 0.0 and props['engHigh'] == 0.0:
            props['engHigh'], props['engLow'] = None, None
    except: pass

    try:
        props['alarm'] = pvControl.metadata.severity 
        #_printd(f'status {pvControl.metadata.severity}')
        if props['alarm'] == 17:# UDF
            props['alarm'] = None
    except: pass

    try:    # legalValues
        enum_strings = pvControl.metadata.enum_strings
        props['legalValues'] = [i.decode() for i in enum_strings]
        props['value'] = props['legalValues'][val]
    except:
        #props['legalValues'] = None
        if 'legalValues' in props:
            del props['legalValues']
    #_printd(f'_props {props}')
    _PVCache[pvName][_PVC_Props] = props

def info(devPar:tuple):
    """Abridged PV info"""
    pvName = _pvname(devPar)
    pv = _get_pv(pvName)
    return {devPar:_PVCache[pvName][_PVC_Props]}

def get(devPar:tuple, *args, **kwargs):
    '''Returns devPar-keyed map of PV properties: {'value','timestamp'...}'''
    pvName = _pvname(devPar)
    #print(f'>get: {devPar, pvName}')
    pv = _get_pv(pvName)
    pvData = pv.read(data_type='time')
    rDict = _unpack_ReadNotifyResponse(pvName, pvData)
    return rDict

def set(devParValue:tuple):
    '''Sets the PV value. The devParValue is (deviceName, ParameterName, value)
    '''
    dev, par, value = devParValue
    #print(f'epicsAccess.set({dev,par,value})')
    pvName = _pvname((dev,par))
    pv = _get_pv(pvName)
    try: # if PV has legalValues then the value should be index of legalValues
        value = _PVCache[pvName][_PVC_Props]['legalValues'].index(value)
    except Exception as e:
        #print(f'in epicsAccess.set. Value not in legalValues: {e}')
        pass
    pv.write(value)
    return 1

def _unpack_ReadNotifyResponse(pvName:str, pvData):
    val = pvData.data
    if len(val) == 1:
        try:    #it as numpy
            val = pvData.data[0].item()
        except: # it is not numpy
            val = pvData.data[0]
    #_printd(f'pvData:{pvData}')
    #_printd(f'val:{val}, {pvData.data_type}')
    if pvData.data_type == _CA_data_type_STRING:
        val = val.decode()

    legalValues = _PVCache[pvName][_PVC_Props].get('legalValues')
    if legalValues is not None:
        #rDict['value'] = legalValues[int(val)]
        val = legalValues[int(val)]
    alarm = pvData.metadata.severity
    key = _PVCache[pvName][_PVC_DevPar]
    rDict = {key: {'value':val\
    , 'timestamp':pvData.metadata.timestamp, 'alarm': alarm}}
    return rDict

def _callback(subscription, pvData):
    #tMark = [timer(), 0., 0.]
    pvName = subscription.pv.name
    _printd(f'>epicsAccess._callback: {pvName}')#{pvData})')
    rDict = _unpack_ReadNotifyResponse(pvName, pvData)
    #tMark[1] = timer()
    cache = _PVCache.get(pvName)
    cb = cache[_PVC_CB]
    if cb:
        cb(rDict)
    #tMark[2] = timer() - tMark[1]
    #tMark[1] -= tMark[0]
    ##_printd(f'caproto cb times {tMark}')# 20-30 uS
    
def subscribe(callback:callable, devPar:tuple):
    '''Subscribe callback method to changes of PVs, defined by devPar'''
    pvName = _pvname(devPar)
    pv = _get_pv(pvName)
    subscription = pv.subscribe(data_type='time')
    _PVCache[pvName][_PVC_CB] = callback
    subscription.add_callback(_callback)
    _Subscriptions.append(subscription)

def unsubscribe():
    '''Unsubscribe all subscriptions'''
    global _Subscriptions
    for subscription in _Subscriptions:
        #print(f'>epicsAccess clear subs: {subscription}')
        subscription.clear()
    _Subscriptions = []
    



#!/usr/bin/env python3
"""This script provide conversion of LDO name to (host:port,device).
Usage:
  from liteserver import liteCNS
  hostname = liteCNS.hostPort('pi108')
"""
# If CNSHostPort is defined, then the name will be resolved using a dedicated
__version__='v02 2023-03-17'# use python-based configuration

import sys

CNSHostPort = None#'hostName;9699'# host;port of the liteServer

#`````````````````````````````````````````````````````````````````````````````
def hostPort(cnsName='*', server=CNSHostPort):
    if server is not None:
        # CNSHostPort is defined, use liteCNSServer
        raise NameError('ERROR:CNS, liteCNSServer is not supported yet')

    #``````````read conguration file into the config dictionary```````````````
    configDir = '/operations/app_store/liteServer'
    moduleName = 'liteCNS_resolv'
    sys.path.append(configDir)
    from importlib import import_module#, reload
    #print(f'importing {moduleName}')
    ConfigModule = import_module(moduleName)
    #print(f'Imported: {configDir}/{moduleName}')
    if cnsName == '*':
        r = ConfigModule.hosts
    else:
        r = ConfigModule.hosts[cnsName]
    return r



#!/usr/bin/env python3
"""Dynamic plotting tool for control system frameworks.
Frameworks supported: EPICS, ADO, LiteServer.
The framework is distinguished by adding a prefix to Process Variable names:
'E:' for EPICS CA, 'V:' for EPICS PVA, 'L:' for LiteServer.
"""
import sys,os
from . import pvplot

def main():
    import argparse
    global pargs
    parser = argparse.ArgumentParser(description=__doc__
    ,formatter_class=argparse.ArgumentDefaultsHelpFormatter
    ,epilog=(f'pvplot: {pvplot.__version__}'))
    parser.add_argument('-a', '--prefix', default='', help=\
      "Prefix, which will be added to all data objects, e.g: 'L:localhost:dev1:'")
    #TODO:parser.add_argument('-H', '--histogram', help=\
    #  'Plot the space-separated parameters as histograms')
    parser.add_argument('-c','--configDir',default='/operations/app_store/pvplot',
      help='Configuration directory')
    parser.add_argument('-f', '--file', help=
      'Configuration file')
    parser.add_argument('-l', '--limit', type=int, default=1048576,
      help='Upper limit for number of points to store')
    parser.add_argument('-s', '--sleepTime', type=float, default=0.1,
      help='sleep time between data delivery [s]')
    parser.add_argument('-v', '--verbose', action='count', default=0, help=\
      'Show more log messages (-vv: show even more).')
    parser.add_argument('-x', '--xscale', help=\
     'Parameter, which provides dynamic scale for X-axis')
    parser.add_argument('-X', '--xrange', help=\
     'Fixed range of X axis, e.g: -x10:20')
    parser.add_argument('-y', '--yscale', help=\
     'Parameter, which provides dynamic scale for Y-axis')
    parser.add_argument('-Y', '--yrange', help=\
     'Fixed range of Y axis, e.g: -y1000:4095')
    parser.add_argument('-z', '--zoomin', type=float, default=1., help=\
      'Zoom the application window by a factor')
    parser.add_argument('-#','--dock', action='append', nargs='*', 
      help='''Plot a specified dock, i.e: to plot two scrolling plots in lower 
        dock and correlation plot in upper dock:
        pvplot -a'L:localhost:dev1:' -#0'yMax yMin' -#1'yMax,yMin' ''')
    parser.add_argument('parms', nargs = '?', default='yMax yMin',
      help='''For stripchart: space-separated list, for correlation plot: comma-separated list of data objects.''')
    pargs = parser.parse_args()
    PVPlot = pvplot.PVPlot
    PVPlot.pargs = pargs

    #``````````````arrange keyboard interrupt to kill the program from terminal.
    import signal
    signal.signal(signal.SIGINT, signal.SIG_DFL)
 
    PVPlot.start()

if __name__ == "__main__":
    main()



"""Module for accessing multiple Process Variables, served by a liteServer.
"""
__version__ = '3.3.0 2024-07-19'#receive_dictio split in two parts
#TODO: Right now the nonblocked _receive_socket is called from subscribtion thread and from channel.transaction(). Are they thread safe? 
#TODO: Recover from timeout. It is tricky. The timeout could be due to slow or stopped server, in that case do not recover.

import sys, time, socket
from os import getpid
import getpass
_timer = time.perf_counter
import threading
receive_dictio_lock = threading.Lock()

# object encoding
#import ubjson
#encoderDump = ubjson.dumpb
#encoderLoad = ubjson.loadb
#import msgpack as encoder
import cbor2 as encoder
encoderDump = encoder.dumps
encoderLoad = encoder.loads

#````````````````````````````Globals``````````````````````````````````````````
Port = 9700
PrefixLength = 4
SocketSize = 1024*64 # max size of UDP transfer
Dev,Par = 0,1
NSDelimiter = ':'# delimiter in the name field

Username = getpass.getuser()
Program = sys.argv[0]
PID = getpid()
def get_user():
    print(f'liteAcces user:{Username}, PID:{PID}, program:{Program}')
#,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,
#````````````````````````````Helper functions`````````````````````````````````
MaxPrint = 500
def _croppedText(obj, limit=200):
    txt = str(obj)
    if len(txt) > limit:
        txt = txt[:limit]+'...'
    return txt
def _printTime(): return time.strftime("%m%d:%H%M%S")
def _printi(msg): 
    print(_croppedText(f'INFO.LA@{_printTime()}: '+msg))

def _printw(msg):
    msg = msg = _croppedText(f'WARN.LA@{_printTime()}: '+msg)
    print(msg)
    #Device.setServerStatusText(msg)

def _printe(msg): 
    msg = _croppedText(f'ERROR.LA@{_printTime()}: '+msg)
    print(msg)
    #Device.setServerStatusText(msg)

def _printv(msg):
    if PVs.Dbg > 0: print('Dbg0.LA: '+msg)
def _printvv(msg):
    if PVs.Dbg > 1: print('Dbg1.LA: '+msg)

def _croppedText(txt, limit=200):
    if len(txt) > limit:
        txt = txt[:limit]+'...'
    return txt

ReceiverStatistics = {'records':0, 'acks':0, 'bytes':0., 'retrans':0, 'time':0.}
ReceiverStatisticsLast = ReceiverStatistics.copy()
def testCallback(args):
    global ReceiverStatisticsLast
    dt = 10.
    ct = time.time()
    if ct - ReceiverStatisticsLast["time"] >= dt:
        stat = ReceiverStatistics.copy()
        del stat['time']
        for i in stat:
            stat[i] = (stat[i] - ReceiverStatisticsLast[i])
        print(f'Received in last {dt}s: {stat}')
        ReceiverStatisticsLast = ReceiverStatistics.copy()

def ip_address():
    """Platform-independent way to get local host IP address"""
    return [(s.connect(('8.8.8.8', 53)), s.getsockname()[0], s.close())\
        for s in [socket.socket(socket.AF_INET, socket.SOCK_DGRAM)]][0][1]

CNSMap = {}# local map of cnsName to host:port
def _hostPort(cnsNameDev:tuple):
    """Return host;port of the cnsName,Dev try it first from already 
    registered records, or from the name service"""
    global CNSMap
    if len(cnsNameDev) == 1:
        msg = f'Device name should be a tuple (dev,name), got: {cnsNameDev}'
        #_printe(msg)
        raise NameError(msg)
    cnsName,dev = cnsNameDev
    if isinstance(cnsName,list):
        cnsName = tuple(cnsName)
    try:  
        hp,dev = CNSMap[cnsName]# check if cnsName is in local map
    except  KeyError:
        from . import liteCNS
        _printi(f'cnsName {cnsName} not in local map: {CNSMap}')
        try:
            hp = liteCNS.hostPort(cnsName)
        #except NameError:
        except Exception as e:
            msg = (f'The host name {cnsName} is not in liteCNS: {e}\n'
                f"Trying to use it as is: '{cnsName}'")
            #raise   NameError(msg)
            _printw(msg)
            hp = cnsName
        # register externally resolved cnsName in local map
        hp = hp.split(';')
        hp = tuple(hp) if len(hp)==2 else (hp[0],Port)            
        #_printi('cnsName %s is locally registered as '%cnsName+str((hp,dev)))
        CNSMap[cnsName] = hp,dev
        _printi(f'Assuming host,port: {hp}')
    except ValueError:
        msg = f'Device name wrong: {cnsNameDev}, should be of the form: host:dev'
        #_printe(msg)
        raise NameError(msg)
    h,p = hp
    try:
        h = socket.gethostbyname(h)
    except:
        _printe(f'Could not resolve host name {h}')
        sys.exit(1)
    return h,p

retransmitInProgress = None
def _recvUdp(sock):
    """Receive the chopped UDP data"""
    sockAddr,port = sock.getsockname()
    #print(f'>_recvUdp {port} locked: {recvLock.locked()}')
    #with recvLock:
    global retransmitInProgress
    chunks = {sock:{}}
    tryMore = 5# Max number of allowed lost packets
    ts = _timer()
    ignoreEOD = 3

    def ask_retransmit(offsetSize):
        global retransmitInProgress
        ReceiverStatistics['retrans'] += 1
        retransmitInProgress = tuple(offsetSize)
        cmd = {'cmd':('retransmit',offsetSize)}
        _printi(f'Asking to retransmit port {port}: {cmd}')
        sock.sendto(encoderDump(cmd),addr)
    
    while tryMore:
        try:
            buf, addr = sock.recvfrom(SocketSize)
            ReceiverStatistics["records"] += 1
            ReceiverStatistics["bytes"] += len(buf)
            ReceiverStatistics["time"] = time.time()
            
        except socket.timeout as e:
            msg = f'Timeout in recvfrom {sockAddr,port}'
            # Don not return, raise exception, otherwise the pypet will not recover
            raise TimeoutError(msg)
        if buf is None:
            raise RuntimeError(msg)
        size = len(buf) - PrefixLength
        offset = int.from_bytes(buf[:PrefixLength],'big')# python3
        
        #DNP_printi(f'chunk received at port {port}: {offset,size}')

        if size > 0:
            chunks[sock][offset,size] = buf[PrefixLength:]
        if offset > 0 and not retransmitInProgress:
            # expect more chunks to come
            continue

        # check if chunk is EOD, i.e. offset,size = 0,0
        if size == 0:
            ignoreEOD -= 1
            if ignoreEOD >= 0:
                #print(f'premature EOD{ignoreEOD} received from, ignore it')
                continue
            else:
                msg = f'Looks like first chunk is missing at port {port}'
                _printw(msg)
                #This is hard to recover. Give up
                return [],0
                #return ('WARNING: '+msg).encode(), addr
                # sortedKeys = sorted(chunks[sock])
                # offsetSize = [0,sortedKeys[0][0]]
                # allAssembled = False
                # ask_retransmit(offsetSize)
                # break
        else:
            #print('First chunk received')
            pass

        if retransmitInProgress is not None:
            if (offset,size) in chunks[sock]:
                _printi(f'retransmission received {offset,size}')
            else:
                _printw(f'server failed to retransmit chunk {retransmitInProgress}')
                tryMore = 1
            retransmitInProgress = None

        # last chunk have been received, offset==0, size!=0
        # check for lost  chunks
        sortedKeys = sorted(chunks[sock])
        prev = [0,0]
        allAssembled = True
        for offset,size in sortedKeys:
            #print('check offset,size:'+str((offset,size)))
            last = prev[0] + prev[1]
            if last != offset:
                l = offset - last
                if l > 65536:
                    msg = f'Lost too many bytes at port {port}: {last,l}, data discarded'
                    _printw(msg)
                    #raise RuntimeError(msg)
                    return [],0
                    #return 'WARNING: '+msg, addr
                ask_retransmit((last, l))
                allAssembled = False
                break
            prev = [offset,size]

        if allAssembled:
            break
        #print(f'tryMore: {tryMore}')
        tryMore -= 1
    ts1 = _timer()
        
    if not allAssembled:
        msg = 'Partial assembly of %i frames'%len(chunks[sock])
        #raise BufferError(msg)
        _printw(msg)
        return [],0
        #return ('WARNING: '+msg).encode(), addr

    data = bytearray()
    sortedKeys = sorted(chunks[sock])
    for offset,size in sortedKeys:
        # _printv('assembled offset,size '+str((offset,size)))
        data += chunks[sock][(offset,size)]
    tf = _timer()
    # if len(data) > 500000:
        # _printv('received %i bytes in %.3fs, assembled in %.6fs'\
        # %(len(data),ts1-ts,tf-ts1))
    #_printi('assembled %i bytes'%len(data))
    return data, addr

def _send_dictio(dictio, sock, hostPort:tuple):
    """low level send"""
    global LastDictio
    LastDictio = dictio.copy()
    dictio['username'] = Username
    dictio['program'] = Program
    dictio['pid'] = PID
    _printv(f'send_dictio to {hostPort}: {dictio}')
    encoded = encoderDump(dictio)
    sock.sendto(encoded, hostPort)

def _send_cmd(cmd, devParDict:dict, sock, hostPort:tuple, values=None):
    import copy
    #_printv(f'_send_cmd({cmd}.{devParDict} to {sock}')
    dpd = devParDict
    if cmd == 'set':
        if len(devParDict) != 1:
            raise ValueError('Set is supported for single device only')
        dpd = copy.deepcopy(devParDict)# that is IMPORTANT! 
        for key in dpd:
            dpd[key] += 'v',values
    devParList = list(dpd.items())
    dictio = {'cmd':(cmd,devParList)}
    #_printv('sending cmd: '+str(dictio))
    _send_dictio(dictio, sock, hostPort)

def _receive_socket(sock, ackHostPort:tuple):
    data, addr = _recvUdp(sock)
    # acknowledge the receiving
    if Access.dbgDrop_Ack > 0:
        Access.dbgDrop_Ack -= 1
    else:
        try:
            sock.sendto(b'ACK', ackHostPort)
            ReceiverStatistics['acks'] += 1
        except OSError as e:
            _printw(f'OSError: {e}')
        # _printv(f'ACK sent to {hostPort}')
        #self.sock.sendto(b'ACK', self.hostPort)
        #_printv('ACK2 sent to '+str(self.hostPort))
    return data, addr

def _decode_data(data, addr):
    """Receive and decode message from associated socket"""
    #if receive_dictio_lock.locked():
    #    _printv('receive_dictio locked')
    #with receive_dictio_lock:

    _printv('received %i bytes'%(len(data)))
    #_printv('received %i of '%len(data)+str(type(data))+' from '+str(addr)':')
    # decode received data
    # allow exception here, it will be caught in execute_cmd
    if len(data) == 0:
        _printw(f'empty reply for: {LastDictio}')
        return {}
    try:
        decoded = encoderLoad(data)
    except Exception as e:
        _printw(f'exception in encoder.load Data[{len(data)}]: {e}')
        #print(str(data)[:150])
        #raise ValueError('in _receive_dictio: '+msg)
        return {}
    #for key,val in decoded.items():
    #    print(f'received from {key}: {val.keys()}')

    if not isinstance(decoded,dict):
        #print('decoded is not dict')
        return decoded
    # JSON does not allow tuples as a keys, therefore the key is a combined string: hos:dev:par
    # Split the key strings to tuple ('host:dev','par).
    parDict = {}
    for (key, value) in decoded.items():
        key = tuple(key.rsplit(':',1))
        if len(key) == 1:
            key = key[0]
        parDict[key] = value
    for parName,item in list(parDict.items()):
        # items could by numpy arrays, the following should decode everything:
        # _printv(f'parName {parName}: {parDict[parName].keys()}')
        # check if it is numpy array
        shapeDtype = parDict[parName].get('numpy')
        if shapeDtype is not None:
            #print(f'par {parName} is numpy {shapeDtype}')
            shape,dtype = shapeDtype
            v = parDict[parName]['value']
            # it is numpy array
            from numpy import frombuffer
            parDict[parName]['value'] =\
                frombuffer(v,dtype).reshape(shape)
            del parDict[parName]['numpy']
        else:
            #print(f'not numpy {parName}')
            pass
    #print(f'\ndecoded: {parDict}')
    #print('<receive_dictio')
    return parDict

class Subscriber():
    def __init__(self, hostPort:tuple, devParDict:dict, callback):
        self.name = f'{hostPort,devParDict}'
        self.hostPort = hostPort
        self.devParDict = devParDict
        self.callback = callback

class SubscriptionSocket():
    event = threading.Event()
    '''handles all subscriptions to single hostPort'''
    def __init__(self, hostPort, sock):
        #_printi(f'>subsSocket {hostPort}')
        self.name = f'{hostPort}'
        self.hostPort = tuple(hostPort)
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.callback = None# holds the callback for checking
        dispatchMode = 'Thread'
        if dispatchMode == 'Thread':
            self.selector = None
            self.socket.settimeout(20)
            self.thread = threading.Thread(target=self.receivingThread)
            self.thread.daemon = True
            self.event.clear()
            self.thread.start()

    def receivingThread(self):
        _printi(f'>receiving thread started for {self.hostPort}')
        while not self.event.is_set():
            try:
                #_printv(f'>subscription receive_dict {self.socket}')
                try:
                    da = _receive_socket(self.socket, self.hostPort)
                    dictio = _decode_data(*da)
                    _printv(f'subs received: {dictio}')
                except TimeoutError as e:
                    msg = f'Timeout in subscription thread: {e}'
                    _printw(msg)
                    continue
            except Exception as e:
                msg = f'in subscription thread socket {self.name}: '+str(e)
                _printw(msg)
                #raise
                sys.exit()
            self.dispatch(dictio)
        _printi(f'<receiving thread stopped for {self.hostPort}') 
        
    def subscribe(self, subscriber):
        if self.socket is None:
            print(f'UDP socket created for {self.hostPort}')
            self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        # just checking:
        if subscriber.hostPort != self.hostPort:
            _printe(f'Subscribe logic error in {self.name}')# this should never happen
        #_printv(f'subscribing {subscriber.name}')
        if self.callback is None:
            self.callback = subscriber.callback
        else:
            if self.callback != subscriber.callback:
                _printe(f'Only one callback is supported per hostPort, subscription for {subscriber.name} is discarded')
                return
        self.event.clear()
        _send_cmd('subscribe', subscriber.devParDict, self.socket\
        , self.hostPort)

    def unsubscribe_all(self):
        _printi(f'unsubscribing {self.hostPort}, {self.socket}')
        if self.socket is None:
            return
        _send_cmd('unsubscribe', {'*':'*'}, self.socket, self.hostPort)
        #_printi(f'killing thread of {self.name}')
        self.event.set()
        #self.thread.raise_exception()
        #print(f'shutting down {self.hostPort}')
        try:    self.socket.shutdown(socket.SHUT_RDWR)
        except Exception as e: 
            pass 
            #_printw(f'Exception in shutting down: {e}')
        #print(f'closing down {self.hostPort}')
        try:    self.socket.close()
        except Exception as e:
            _printw(f'Exception in closing: {e}')
        self.socket = None

    def dispatch(self, dictio):
        if dictio:
            #_printi(_croppedText(f'>dispatch {dictio}'))
            self.callback(dictio)
        else:
            _printw(f'empty data from {self.hostPort}')
            #if self.selector:
            #    self.selector.unregister(self.socket)
            #self.socket.close()
            return
    
subscriptionSockets = {}

def _add_subscriber(hostPort:tuple, devParDict:dict, sock, callback=testCallback):
    subscriber = Subscriber(hostPort, devParDict, callback)
    # register new socket if not registered yet
    subsSocket = subscriptionSockets.get(hostPort)
    if subsSocket is None:
        print(f'adding new subscriber {hostPort}')
        subsSocket = SubscriptionSocket(hostPort, sock)
        subscriptionSockets[hostPort] =  subsSocket
        #_printv(f'new socket in publishingHouse: {hostPort}')
    subscriptionSockets[hostPort] = subsSocket
    subsSocket.subscribe(subscriber)

def unsubscribe_all():
    global subscriptionSockets
    if len(subscriptionSockets) > 0:
        for hostPort,subsSocket in subscriptionSockets.items():
            subsSocket.unsubscribe_all()
        subscriptionSockets = {}
        _printi('All unsubscribed')

pvSockets = {}
class Channel():
    Perf = False
    """Provides access to host;port.
    If a socket for host;port have been opened before, it will be reused."""
    def __init__(self, hostPort:tuple, devParDict={}, timeout=10):
        _printv(f'>Channel {hostPort,devParDict}')
        self.devParDict = devParDict
        host = hostPort[0]
        if host.lower() in ('','localhost'):
            host = ip_address()
        try:    port = int(hostPort[1])
        except: port = 9700
        self.hostPort = host,port
        self.timeout = timeout
        self.name = f'{self.hostPort}'
        self.recvMax = 1024*1024*4
        _printv('Try to reuse existing socket')
        self.sock = pvSockets.get(self.hostPort)
        if self.sock is None:
            print('There is no sockets for that host, create a new socket')
            self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            self.sock.settimeout(timeout)
            #self.sock.setblocking(False)
            #events = selectors.EVENT_READ | selectors.EVENT_WRITE
            #Sel.register(sock, events, data=message)
            pvSockets[self.hostPort] = self.sock
        _printv(f'<Channel {hostPort,devParDict}: {self.sock}')

    def _transaction(self, cmd, value=None):
        # normal transaction: send command, receive response
        if Channel.Perf:
            ts = _timer()
        _send_cmd(cmd, self.devParDict, self.sock, self.hostPort, value)
        #r = True if cmd == 'set' else _receive_dictio(self.sock, self.hostPort)
        da = _receive_socket(self.sock, self.hostPort)
        r = _decode_data(*da)
        if Channel.Perf: print('transaction time: %.5f'%(_timer()-ts))
        return r
    
class PVs(object): #inheritance from object is needed in python2 for properties to work
    """Class, representing multiple data access objects."""
    Dbg = 0
    subscriptionsCancelled = True
    #Cache = {}

    def __init__(self, *ldoPars):
        # unpack arguments to hosRequest map
        self.channelMap = {}
        if isinstance(ldoPars[0], str):
            _printe('Device,parameter should be a list or tuple')
            return
            sys.exit(1)
        for ldoPar in ldoPars:
            #if ldoPar in PVs.Cache:
            #    _printv(f'ldoPar {ldoPar} is already exist')
            _printv(f'``````````````````Instantiating PVs ldoPar:{ldoPar}')
            ldo = ldoPar[0]
            if len(ldoPar) == 1:
                pars = '*'
            else:
                if isinstance(ldoPar[1],str):
                    ldoPar = list(ldoPar)
                    ldoPar[1] = [ldoPar[1]]
                pars = ldoPar[1:]
            try:    ldo = ldo.split(NSDelimiter)
            except: pass
            ldo = tuple(ldo)
            #if isinstance(pars,str): pars = [[pars]]
            # ldo is in form: (hostName,devName)
            ldoHost = _hostPort(ldo)
            cnsNameDev = NSDelimiter.join(ldo)
            if ldoHost not in self.channelMap:
                self.channelMap[ldoHost] = {cnsNameDev:pars}
                #_printv(f'created self.channelMap[{ldoHost,self.channelMap[ldoHost]}')
            else:
                if False:#try:
                    _printv(f'try to append old cnsNameDev {ldoHost,cnsNameDev} with {pars[0]}')
                    self.channelMap[ldoHost][cnsNameDev][0].append(pars[0])
                else:#except:
                    _printv(f'creating new cnsNameDev {ldoHost,cnsNameDev} with {pars[0]}')
                    self.channelMap[ldoHost][cnsNameDev] = pars
                print(f'updated self.channelMap[{ldoHost}: {self.channelMap[ldoHost]}')
        channelList = list(self.channelMap.items())
        _printv(f',,,,,,,,,,,,,,,,,,,channelList constructed: {channelList}')
        self.channels = [Channel(*i) for i in channelList]
        return

    def info(self):
        for channel in self.channels:
            return channel._transaction('info')

    def get(self):
        for channel in self.channels:
            return channel._transaction('get')

    def _firstValueAndTime(self):
        try:
            firstDict = self.channels[0]._transaction('get')
            if not isinstance(firstDict,dict):
                return firstDict
            firstValsTDict = list(firstDict.values())[0]
        except Exception as e:
            _printw('in _firstValueAndTime: '+str(e))
            return (None,)
        t = firstValsTDict.get('timestamp')
        v = firstValsTDict.get('value')
        return v,t

    #``````````````Property 'value````````````````````````````````````````````
    # It is for frequently needed get/set access to a single parameter
    @property
    def value(self):
        """Request from server first item of the PVs and return its 
        value and timestamp,"""
        return self._firstValueAndTime()

    @value.setter
    def value(self,value):
        """Send command to set the value to the first item of the PVs"""
        return self.set(value)
    #,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,

    def read(self):
        """Return only readable parameters"""
        for channel in self.channels:
            return channel._transaction('read')

    def set(self,value):
        """On normal completion returns replies from channels.
        On error - raises RuntimeError."""
        r = {}
        for channel in self.channels[:1]:
            _printv(f'>set {channel.devParDict} {value}')
            reply = channel._transaction('set',value)
            _printv(f'<set: {reply}')
            if isinstance(reply,str):
                raise RuntimeError(reply)
            r.update(reply)
        return r

    #``````````````subscription ``````````````````````````````````````````````
    def subscribe(self, callback=testCallback):
        if len(self.channels) > 1:
            raise NameError('subscription is supported only for single host;port')
        channel = self.channels[0]
        _add_subscriber(channel.hostPort, channel.devParDict, channel.sock, callback)

    def unsubscribe(self):
        _printi(f'>PV.insubscribe: {self.channels[:1]}')
        unsubscribe_all()

#``````````````````Universal Access```````````````````````````````````````````
#PVC_PV, PVC_CB, PVC_Props = 0, 1, 2
PVC_PV, PVC_Thread = 0, 1
    
class Access():
    """Interface to Process Variables, served by liteServer.
    The pvName should be a tuple: (deviceName,parameterName)
    The full form of the device name: hostName;Port:deviceName,
    if Port is default then it can be omitted: hostName:deviceName.
    Returned values are remapped to the form {(hostDev,par):{parprop:{props}...}...}
    that could be time consuming for complex requests.
    """
    _Subscriptions = []
    __version__ = __version__
    dbgDrop_Ack = 0# Debugging. Number of ACK to drop.

    def set_dbg(dbg):
        PVs.Dbg = dbg

    def info(*devParNames):
        return PVs(*devParNames).info()

    def get(*devParNames, **kwargs):# kwargs are for compatibility with older versions
        return PVs(*devParNames).get()

    def set(dev_pars_values):
        dev,pars,values = dev_pars_values
        return PVs((dev,pars)).set(values)

    def subscribe(callback, *devParNames):
        if not callable(callback):
            _printe(('subscribe arguments are wrong,'
            'expected: subscribe(callback,(dev,par))'))
            return
        pvs = PVs(*devParNames)
        pvs.subscribe(callback)

    def unsubscribe():
        """Unsubscribe all parameters"""
        if len(subscriptionSockets) > 0:
            _printi('>Access.unsubscribe()')
            unsubscribe_all()
#,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,



#Pypet page to control a Peak Simulator in LiteServer infrastructure.

#``````````````````Variables, used by pypet
_Namespace = "LITE"# Devices are from Liteserver infrastructure
_WindowTitle = 'LITE PeakSimulator'
_PageHelp = 'https://github.com/ASukhanov/liteServer/blob/master/liteserver/device/litePeakSimulator.py'

#``````````````````Definitions`````````````````````````````````````````````````

hostPort = "acnlin23;9710:"
#hostPort = "localhost;9710:"
dev = hostPort+"dev1:"
server = hostPort+"server:"
bkgColor = {"color": (240,240,240)}# background color
_ = {'':bkgColor}# empty cell

# Python expressions and functions, used in the spreadsheet.
def span(x,y): return {'span':[x,y]}
def color(*v): return {'color':v[0]} if len(v)==1 else {'color':list(v)}
def font(size): return {'font':['Arial',size]}
def slider(minValue,maxValue):
    """Definition of the GUI element: horizontal slider with flexible range"""
    return {'widget':'hslider','opLimits':[minValue,maxValue]}

# Buttons
def Button(name, launch, fontSize=12, col='lightGreen'):
    return {name:{'launch':launch, 'font':['Open Sans Semibold',fontSize], 
    **color(col)}}
PlotXY = Button("Plot Y vs X",
  "python3 -m pvplot -s.01 -a'L:localhost;9710:dev1:' 'x,y'")
Stripchart = Button("Stripchart",
  "python3 -m pvplot -s.01 -a'L:localhost;9710:dev1:' 'yMin yMax'")
Correlation = Button("Min-Max Correlation",
  "python3 -m pvplot -s.01 -a'L:localhost;9710:dev1:' 'yMin,yMax'")

GrayRow = {'ATTRIBUTES':{'color':'lightGray', 'justify':'center', **font(12)}}
LargeRow = {'ATTRIBUTES':{**font(12)}}
FocusRow = {'ATTRIBUTES':{'color':'lightCyan', **font(12)}}
#,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,
#``````````````````Page attributes, optional``````````````````````````````````
#_Page = {'editable':False, **color(252,252,237)}
_Page = {**bkgColor}

_Columns = {
  1: {"justify": "right"},#"center"},
  2: {"width": 100},
  3: {"width": 60, "justify": "right"},
}

#``````````````````Rows in the spreadsheet`````````````````````````````````````
_Rows = [
["Device is ", dev+"run", 'debug:', server+'debug'],
["Device status messages:", {dev+"status":span(3,1)}],
["Data update frequency:", dev+"frequency", "nPoints:", dev+"nPoints"],
[],
[GrayRow,{'Plots and data':span(4,1)}],
[Correlation,Stripchart,'',PlotXY,],
["X:", {dev+"x":span(3,1)}],
["Y:", {dev+"y":span(3,1)}],
['yMin:', dev+'yMin', 'yMax:', dev+'yMax'],
[],
[GrayRow,{'Peak Parameters':span(4,1)}],
["Noise level:", {dev+"noise":slider(0.,20.)}, dev+"noise",_],
["Swinging range of peaks:", {dev+"swing":slider(0.,2.)}, dev+"swing",_],
["Peaks parameters:", {dev+"peakPars":span(3,1)}],
["Quadratic baseline:", {dev+"background":span(3,1)}],
[],
[GrayRow,{'For experts, performance':span(4,1)}],
['Array of perf. counters:', {server+'perf':span(3,1)}],
['MBytes published:', server+'perf[1]', 'MB/s:', server+'perf[2]'],
['Packets lost,dropped:', server+'perf[4:6]',_,_],
["Cycles:", dev+"cycle", "Cycle/s", dev+"rps"],
]
